#! /usr/bin/python3
# -*- coding: utf-8 -*-
# @FileName: template.py
# @Author: Li Chengxin 
# @Time: 2023/6/30 22:26


REGIONS_TMP = {
    "shape_attributes":
        {
            "name": "polygon",
            "all_points_x": [],
            "all_points_y": []
        },
    "region_attributes":
        {
            "cell_type": None,
            "cell_id": None,
            'track_id': None,
        }
}

FRAME_TMP =  {
        "filename": None,
        "size": None,
        "regions": [],
        "file_attributes": {}
    }



from __future__ import annotations

import os
from functools import lru_cache
import math
from typing import Tuple, List
import cv2
import json

import matplotlib.pyplot as plt
import numpy as np
import shapely.errors
from tifffile import tifffile

from SCTrack import config
from SCTrack.base import Cell, Vector


class Mask(object):
    def __init__(self, mask=None, center=None, coord=None):
        self._mask = mask
        self._center = center
        self.id = None
        self.frame_index = None
        self.coord = coord

    @property
    def mask(self):
        if self._mask is None:
            raise ValueError("No mask")
        return self._mask

    @property
    def center(self) -> Tuple[int | float]:
        if self._center is None:
            raise ValueError("No available center position")
        return self._center

    def __str__(self):
        return f"Mask object at center of {self._center[0], self._center[1]}"

    def __repr__(self):
        return self.__str__()


class Feature(object):
    """
    All features contained in each cell instance, including the following:
     - area: cell area
     - bbox_area: bounding box area
     - shape: sequence of cell outline coordinates
     - center: cell center coordinates
     - vector: the vector of the cell relative to the origin
     - bbox: the bounding box coordinates of the cell [y_min, y_max, x_min, x_max]
     - dic_intensity: the dic gray value intensity of the mask area
     - mcy_intensity: mcy gray value intensity of the mask area
     - cell_type forecast period


    """

    def __init__(self, center, bbox, area=None, shape=None, cell_type=None,
                 dic_intensity=None, dic_variance=None, mcy_intensity=None, mcy_variance=None, frame=None):
        self.center = center
        self.bbox = bbox
        self.cell_type = cell_type
        self.mcy_intensity = mcy_intensity
        self.mcy_variance = mcy_variance
        self.dic_intensity = dic_intensity
        self.dic_variance = dic_variance
        self.shape = shape
        self.area = area
        self.vector = Vector(*center)
        self.frame_id = frame


class FeatureExtractor(object):
    """Extract available features for each cell in a single image"""
    _instances = {}

    def __new__(cls, *args, **kwargs):
        key = str(args) + str(kwargs)
        if key not in cls._instances:
            cls._instances[key] = super(FeatureExtractor, cls).__new__(cls)
            cls._instances[key]._init_cell_flag = False
            cls._instances[key]._init_flag = False
            cls._instances[key].__cells = None
        return cls._instances[key]

    def __init__(self, image_dic: np.ndarray | None = None, image_mcy: np.ndarray | None = None,
                 annotation: dict = None,
                 *args, **kwargs):
        """
        image_dic: np.ndarray dic image information 2048x2048
        image_mcy: np.ndarray mcy image information 2048x2048
        annotation: dict Annotation information such as the outline and cycle of cells, regions in the json file
        """
        if not self._init_flag:
            # config.USING_IMAGE_FOR_TRACKING = config.USING_IMAGE_FOR_TRACKING
            self.frame_id = None
            if (image_mcy is not None) and (image_dic is not None):
                if type(image_mcy) != np.uint8:
                    self.mcy = self.convert_dtype(image_mcy)
                else:
                    self.mcy = image_mcy
                if type(image_dic) != np.uint8:
                    self.dic = self.convert_dtype(image_dic)
                else:
                    self.dic = image_dic
            else:
                # config.py configures using image, but the parameters are not compliant, turn off this option
                if config.USING_IMAGE_FOR_TRACKING is True:
                    config.USING_IMAGE_FOR_TRACKING = False
            self.annotation = annotation
            # self.image_shape = self.mcy.shape
            self.frame_index = kwargs.get('frame_index')
            self._init_flag = True

    def coord2counter(self, coord):
        points = []
        for j in range(len(coord[0])):
            x = int(coord[0][j])
            y = int(coord[1][j])
            points.append((x, y))
        contours = np.array(points)
        return contours

    def coordinate2mask(self, coords: np.ndarray | list | tuple, shape, value: int = 255) -> List[Mask]:
        """
        Draw the mask according to the contour coordinates. If you only pass in a set of contour coordinate values,
        be sure to put them in the list and pass in the function.
        For example, coord = ([x1 x2 ... xn], [y1 y2 ... yn]), please call it according to coordinate2mask([coord])
        """
        results = []
        for coord in coords:
            mask = np.zeros(shape, dtype=np.uint8)
            points = []
            for j in range(len(coord[0])):
                x = int(coord[0][j])
                y = int(coord[1][j])
                points.append((x, y))
            contours = np.array(points)
            cv2.fillConvexPoly(mask, contours, (value, 0, 0))
            mMask = Mask(mask, center=(round(float(np.mean(coord[0]))), round(float(np.mean(coord[1])))), coord=coord)
            results.append(mMask)
        return results

    def get_roi_from_mask(self, mask, image):
        pass

    def GetAreaByVector(self, points: List):
        # Calculate polygon area based on vector cross product
        area = 0
        if len(points) < 3:
            raise Exception("error")

        for i in range(0, len(points) - 1):
            p1 = points[i]
            p2 = points[i + 1]
            triArea = (p1.x * p2.y - p2.x * p1.y) / 2
            area += triArea
        return abs(area)

    def area(self, cell):
        """cell area"""
        _area = cv2.contourArea(self.coord2counter(cell.position))
        return _area

    def bbox(self, cell: Cell):
        """bounding box coordinates"""
        x0 = math.floor(np.min(cell.position[0])) if math.floor(np.min(cell.position[0])) > 0 else 0
        x1 = math.ceil(np.max(cell.position[0]))
        y0 = math.floor(np.min(cell.position[1])) if math.floor(np.min(cell.position[1])) > 0 else 0
        y1 = math.ceil(np.max(cell.position[1]))
        return y0, y1, x0, x1

    def get_roi_from_coord(self, cell: Cell, image: np.ndarray):
        """
        Use the cell outline to obtain the dic image or the mcy image, depending on the incoming image parameters.
        :param cell: Cell object
        :param image: dic image or mcy image, that is, the parameter self.mcy or self.dic
        :return: roi np.ndarray
        """
        # x0 = int(np.min(cell.position[0]))
        # x1 = math.ceil(np.max(cell.position[0]))
        # y0 = int(np.min(cell.position[1]))
        # y1 = math.ceil(np.max(cell.position[1]))
        y0, y1, x0, x1 = self.bbox(cell)
        return image[y0: y1, x0: x1]

    def ellipse_points(self, center, rx, ry, num_points=100, theta=0):
        all_x = []
        all_y = []
        for i in range(num_points):
            t = i * 2 * np.pi / num_points
            x = center[0] + rx * np.cos(t) * np.cos(theta) - ry * np.sin(t) * np.sin(theta)
            y = center[1] + rx * np.cos(t) * np.sin(theta) + ry * np.sin(t) * np.cos(theta)
            all_x.append(x)
            all_y.append(y)
        return all_x, all_y

    @lru_cache(maxsize=None)
    def get_cell_list(self):
        """
        Get all cells in each single frame image
        """
        cell_list = []
        for region in self.annotation:
            if region['shape_attributes'].get('name') == 'ellipse':
                rx = region['shape_attributes'].get('rx')
                ry = region['shape_attributes'].get('ry')
                cx = region['shape_attributes'].get('cx')
                cy = region['shape_attributes'].get('cy')
                theta = region['shape_attributes'].get('theta')
                phase = region['region_attributes'].get('phase')
                cell_type = region['region_attributes'].get('cell_type')
                cell_type = cell_type if cell_type else phase
                all_x, all_y = self.ellipse_points((cx, cy), rx, ry, num_points=32, theta=theta)
            elif region['shape_attributes'].get('name') == 'polygon':
                all_x = region['shape_attributes']['all_points_x']
                all_y = region['shape_attributes']['all_points_y']
                all_x = [0 if i < 0 else i for i in all_x]
                all_y = [0 if j < 0 else j for j in all_y]
                phase = region['region_attributes'].get('phase')
                cell_type = region['region_attributes'].get('cell_type')
                cell_type = cell_type if cell_type else phase
                # cell_type = None
            else:
                continue
            try:
                cell = Cell(position=(all_x, all_y), cell_type=cell_type, frame_index=self.frame_index)
                cell.set_region(region)
                _ = cell.center
                _ = cell.polygon
                cell_list.append(cell)
            except ZeroDivisionError:
                continue
            except ValueError as e:
                continue
        return cell_list

    def get_cell_image(self, cell: Cell):
        """
        Get dic and mcy images of cells
        """
        if config.USING_IMAGE_FOR_TRACKING:
            dic = self.get_roi_from_coord(cell, self.dic)
            mcy = self.get_roi_from_coord(cell, self.mcy)
            return dic, mcy
        else:
            return None

    def set_cell_image(self, cell: Cell):
        """
        Set dic information and mcy information for cell instance
        :param cell: Cell object
        :return: Cell object containing image information
        """
        if config.USING_IMAGE_FOR_TRACKING:
            dic, mcy = self.get_cell_image(cell)
            cell.dic = dic
            cell.mcy = mcy
        return cell

    @property
    @lru_cache(maxsize=None)
    def _cells(self):
        cells = self.get_cell_list()
        if config.USING_IMAGE_FOR_TRACKING:
            for cell in cells:
                self.set_cell_image(cell)
        return cells

    def extract(self, cell: Cell) -> Feature:
        """
        :param cell: Cell object
        :return: Feature object
        """
        if config.USING_IMAGE_FOR_TRACKING:
            mcy_intensity = np.mean(cell.mcy)
            mcy_variance = np.std(cell.mcy) ** 2
            dic_intensity = np.mean(cell.dic)
            dic_variance = np.std(cell.dic) ** 2
        else:
            mcy_intensity = mcy_variance = dic_intensity = dic_variance = None
        feature = Feature(center=cell.center, bbox=self.bbox(cell), shape=cell.position,
                          mcy_intensity=mcy_intensity, mcy_variance=mcy_variance,
                          dic_intensity=dic_intensity, dic_variance=dic_variance)

        return feature

    @lru_cache(maxsize=None)
    def __cells_(self):
        _cells = []
        for cell in self._cells:
            feature = self.extract(cell)
            cell.set_feature(feature)
            _cells.append(cell)
        return _cells

    @property
    @lru_cache(maxsize=None)
    def cells(self):
        if self._init_cell_flag is False:
            self.__cells = self.__cells_()
            self._init_cell_flag = True
            return self.__cells
        else:
            return self.__cells

    def add_cell(self, new_cell: Cell):
        self.__cells.append(new_cell)

    def convert_dtype(self, __image: np.ndarray) -> np.ndarray:
        """Convert image from uint16 to uint8"""
        min_16bit = np.min(__image)
        max_16bit = np.max(__image)
        image_8bit = np.array(np.rint(255 * ((__image - min_16bit) / (max_16bit - min_16bit))), dtype=np.uint8)
        return image_8bit

    @staticmethod
    def show(image):
        plt.imshow(image, cmap='gray')
        plt.show()

    def get_regions(self):
        return self.annotation

    def __str__(self):
        return f'Feature Extractor of {self.frame_index if self.frame_index else 0} at {id(self)}'

    def __repr__(self):
        return self.__str__()


def imread(filepath: str | os.PathLike) -> np.ndarray:
    return tifffile.imread(filepath)


def get_frame_by_index(image: np.ndarray, index: int) -> np.ndarray:
    return image[index]


def feature_extract(mcy, dic, jsonfile: str | dict):
    """
    Return the FeatureExtractor instance frame by frame,
    including the current frame, the previous frame, and the next frame
    """
    if type(jsonfile) is str:
        with open(jsonfile, encoding='utf-8') as f:
            annotations = json.load(f)
    elif type(jsonfile) is dict:
        annotations = jsonfile
    else:
        raise TypeError(f"type {type(jsonfile)} are not invalid")
    if mcy and dic:
        _dic = imread(dic)
        _mcy = imread(mcy)
        _frame_len = _mcy.shape[0]
        using_image = True
    else:
        _frame_len = len(annotations)
        using_image = False

    def get_fe(frame_index, frame_name, using_image=False):
        if using_image:
            dic_image = get_frame_by_index(_dic, frame_index)
            mcy_image = get_frame_by_index(_mcy, frame_index)
        else:
            dic_image = mcy_image = None
        region = annotations[frame_name.replace('.tif', '.png')]['regions']
        return FeatureExtractor(image_dic=dic_image, image_mcy=mcy_image, annotation=region, frame_index=frame_index)

    def get_base_name(annotation, index):
        return list(annotation.keys())[index]

    for i in range(_frame_len):
        current_frame_index = i
        if i == 0:
            before_frame_index = 0
        else:
            before_frame_index = i - 1
        if i == _frame_len - 1:
            after_frame_index = i
        else:
            after_frame_index = i + 1
        if using_image:
            before_frame_name = os.path.basename(mcy).replace('.tif', '-' + str(before_frame_index).zfill(4) + '.tif')
            after_frame_name = os.path.basename(mcy).replace('.tif', '-' + str(after_frame_index).zfill(4) + '.tif')
            current_frame_name = os.path.basename(mcy).replace('.tif', '-' + str(current_frame_index).zfill(4) + '.tif')
        else:
            before_frame_name = get_base_name(annotations, before_frame_index)
            current_frame_name = get_base_name(annotations, current_frame_index)
            after_frame_name = get_base_name(annotations, after_frame_index)
        before_fe = get_fe(before_frame_index, before_frame_name, using_image=using_image)
        current_fe = get_fe(current_frame_index, current_frame_name, using_image=using_image)
        after_fe = get_fe(after_frame_index, after_frame_name, using_image=using_image)
        yield before_fe, current_fe, after_fe


if __name__ == '__main__':
    pass



from __future__ import annotations

import csv
import os
import sys
import time
from collections import deque
from concurrent.futures import ThreadPoolExecutor
import shapely

sys.path.append('.')
sys.path.append('..')
sys.path.append('../../')
sys.setrecursionlimit(3000)

from functools import wraps, lru_cache
import warnings
from typing import List
import tifffile
import numpy as np
import cv2

import treelib
from treelib import Tree, Node
from tqdm import tqdm

from SCTrack.utils import convert_dtype, readTif
from SCTrack.base import Cell, Rectangle, Vector, TreeStatus, CellStatus
from SCTrack.t_error import InsertError, MitosisError, NodeExistError, ErrorMatchMitosis, StatusError
from SCTrack.feature import FeatureExtractor, feature_extract
from SCTrack import config

TEST = False
TEST_INDEX = None
CELL_NUM = 0


class Checker(object):
    """
    A checker that checks whether the cells participating in the match can calculate.
    """
    _protocols = [None, 'calcIoU', 'calcCosDistance', 'calcCosSimilar', 'calcEuclideanDistance',
                  'compareDicSimilar', 'compareMcySimilar', 'compareShapeSimilar']

    def __init__(self, protocol=None):
        self.protocol = protocol

    def check(self):
        pass

    def __call__(self, method):

        if self.protocol not in self._protocols:
            warnings.warn(f"Don't support protocol: {self.protocol}, now just support {self._protocols}")
            _protocol = None
        else:
            _protocol = self.protocol

        @wraps(method)
        def wrapper(*args, **kwargs):
            """args[0]: object of func; args[1]: param 1 of func method; args[2]: param 2 of func method"""
            # print(_protocol)
            # print('args:', *args)
            if _protocol is None:
                return method(*args, **kwargs)
            elif _protocol == 'calcIoU':
                return method(*args, **kwargs)
            elif _protocol == 'calcCosDistance':
                return method(*args, **kwargs)
            elif _protocol == 'calcCosSimilar':
                return method(*args, **kwargs)
            elif _protocol == 'compareDicSimilar':
                return method(*args, **kwargs)
            elif _protocol == 'compareMcySimilar':
                return method(*args, **kwargs)
            elif _protocol == 'compareShapeSimilar':
                return method(*args, **kwargs)
            else:
                return method(*args, **kwargs)

        return wrapper


class CellNode(Node):
    """
    Tracking node, including the track_ID, cell_id, branch_id of the cell, and the detailed information of the cell,
    the parent-child node relationship
    """
    _instance_ = {}
    STATUS = ['ACCURATE', 'ACCURATE-FL', 'INACCURATE', 'INACCURATE-MATCH', 'PREDICTED']

    def __new__(cls, *args, **kwargs):
        key = str(args) + str(kwargs)
        if key not in cls._instance_:
            cls._instance_[key] = super().__new__(cls)
            cls._instance_[key].status = None
            cls._instance_[key].track_id = None
            cls._instance_[key].__branch_id = None
            cls._instance_[key].parent = None
            cls._instance_[key].childs = []
            cls._instance_[key].add_tree = False  # if adding to the TrackingTree，set as True
            cls._instance_[key].life = 5
            cls._instance_[key]._init_flag = False
        return cls._instance_[key]

    def __init__(self, cell: Cell, node_type='cell', fill_gap_index=None):
        if not self._init_flag:
            self.cell = cell
            if node_type == 'gap':
                assert fill_gap_index is not None
            super().__init__(cell)
            self._init_flag = True

    @property
    def identifier(self):
        """The unique ID of a node within the scope of a TrackingTree"""
        return str(id(self.cell))

    @property
    def nid(self):
        """same as identifier, override treelib related methods"""
        return self.identifier

    def _set_identifier(self, nid):
        if nid is None:
            self._identifier = str(id(self.cell))
        else:
            self._identifier = nid

    def get_status(self):
        return self.status

    def set_parent(self, parent: CellNode):
        self.parent = parent

    def get_parent(self):
        return self.parent

    def set_children(self, child: CellNode):
        self.childs.append(child)

    def get_children(self):
        return self.childs

    def set_tree_status(self, status: TreeStatus):
        self.tree_status = status
        self.add_tree = True

    def get_tree_status(self):
        if self.add_tree:
            return self.tree_status
        return None

    def set_status(self, status):
        if status in self.STATUS:
            self.status = status
        else:
            raise ValueError(f"set error status: {status}")

    def get_branch_id(self):
        return self.__branch_id

    def set_branch_id(self, branch_id):
        self.__branch_id = branch_id
        self.cell.set_branch_id(branch_id)

    def get_track_id(self):
        if self.track_id is None:
            raise ValueError("Don't set the track_id")
        else:
            return self.track_id

    def set_track_id(self, track_id):
        self.track_id = track_id

    def __repr__(self):
        if self.add_tree:
            return f"Cell Node of {self.cell}, status: {self.get_tree_status()}"
        else:
            return f"Cell Node of {self.cell}"

    def __str__(self):
        return self.__repr__()

    def __hash__(self):
        return int(id(self))


class TrackingTree(Tree):
    """
    The core structure of tracking implement. All tracking results are stored in the instance of this class.
    Each TrackingTree instance represents the cell line of a cell, and the TrackingTree branch represents cell division.
    """

    def __init__(self, root: CellNode = None, track_id=None):
        super().__init__()
        self.root = root
        self.track_id = track_id
        self.mitosis_start_flag = False
        self.status = TreeStatus(self)
        self.__last_layer = []
        self._exist_branch_id = []
        self._available_branch_id = 1
        self.m_counter = 5

    def __contains__(self, item):
        return item.identifier in self._nodes

    def change_mitosis_flag(self, flag: bool):
        """
        When the cell enters mitosis for the first time, self.mitosis_start_flag is set to True,
        and when the cell completes division, it is reset to false
        """
        self.mitosis_start_flag = flag

    def add_node(self, node: CellNode, parent: CellNode = None):
        """Add a new CellNode object to the TrackingTree"""
        node.set_parent(parent)
        if parent is not None:
            parent.childs.append(node)
        super().add_node(node, parent)

    def get_parent(self, node: CellNode):
        """Returns the parent CellNode of a node"""
        return node.get_parent()

    def get_childs(self, node: CellNode):
        """Returns the child CellNode of a node"""
        return node.get_children()

    @property
    def last_layer(self):
        """Return all CellNodes in the last layer of TrackingTree"""
        return self.__last_layer

    @property
    def last_layer_cell(self):
        """Return a dict of {leaf node: cells contained in the node}."""
        cells = {}
        for node in self.leaves():
            cells[node.cell] = node
        return cells

    def update_last_layer(self, node_list: List[CellNode]):
        self.__last_layer = node_list

    def auto_update_last_layer(self):
        self.__last_layer = self.leaves()

    def branch_id_distributor(self):
        """
        Used to assign a branch_id to each branch in a TrackingTree object.
        The branch_id is incremented in a non-reversible and non-reusable manner.
        """
        if self._available_branch_id not in self._exist_branch_id:
            self._exist_branch_id.append(self._available_branch_id)
            self._available_branch_id += 1
            return self._available_branch_id - 1
        else:
            i = 1
            while True:
                if self._available_branch_id + i not in self._exist_branch_id:
                    self._available_branch_id += (i + 1)
                    return self._available_branch_id + i
                i += 1


class Match(object):
    """
    Matcher, matches targets based on previous and next frames and assigns IDs in the current frame.
    Mainly used for feature matching and calculating similarity.
    The unit of operation is frames.
    """
    _instances = {}

    def __new__(cls, *args, **kwargs):
        key = str(args) + str(kwargs)
        if key not in cls._instances:
            cls._instances[key] = super().__new__(cls)
        return cls._instances[key]

    def normalize(self, x, _range=(0, np.pi / 2)):
        """Transforming the value to the range of [0, π/2]."""
        return _range[0] + (_range[1] - _range[0]) * x

    def calcIoU_roughly(self, cell_1: Cell, cell_2: Cell):
        """
        Calculate IoU of two cell bounding boxes.
        return value range: float(0-1)
        """
        rect1 = Rectangle(*cell_1.bbox)
        rect2 = Rectangle(*cell_2.bbox)

        if not rect1.isIntersect(rect2):
            return 0
        elif rect1.isInclude(rect2):
            return 1
        else:
            intersection = Rectangle(min(rect1.x_min, rect2.x_min), max(rect1.x_max, rect2.x_max),
                                     min(rect1.y_min, rect2.y_min), max(rect1.y_max, rect2.y_max))
            union = Rectangle(max(rect1.x_min, rect2.x_min), min(rect1.x_max, rect2.x_max),
                              max(rect1.y_min, rect2.y_min), min(rect1.y_max, rect2.y_max))
            return union.area / intersection.area

    def calcIoU(self, cell_1: Cell, cell_2: Cell):
        """
        Calculate IoU of two cells.
        First, try to calculate the IoU of the contour. If it is not possible to calculate,
        then change to calculating the IoU of the bounding box.
        return value range: float(0-1)
        """
        poly1 = cell_1.polygon
        poly2 = cell_2.polygon

        try:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                intersection = poly1.intersection(poly2)
                union = poly1.union(poly2)
                if union.area == 0:
                    return 0
                elif poly1.contains(poly2) or poly2.contains(poly1):
                    return 1
                else:
                    return intersection.area / union.area
        except shapely.errors.GEOSException:
            return self.calcIoU_roughly(cell_1, cell_2)

    def calcCosDistance(self, cell_1: Cell, cell_2: Cell):
        """
        Calculate the cosine distance of two cells
        Return value range: float[0, 2]
        The smaller the distance, the more similar it is, scaled to [0, 1) by the arctangent function
        The return value is the cosine value after normalization [0, π/2].
        The smaller the return value, the lower the similarity
        """
        dist = cell_1.vector.cosDistance(cell_2.vector)
        return np.cos(np.arctan(dist) / (np.pi / 2))

    def calcCosSimilar(self, cell_1: Cell, cell_2: Cell):
        """
        Calculate the cosine similarity of the center points of two cells
        Return value range: float[0, 1]
        The larger the value, the more similar
        The return value is the sine value after normalization [0, π/2]
        """
        score = cell_1.vector.cosSimilar(cell_2.vector)
        return np.sin(self.normalize(score))

    def calcAreaSimilar(self, cell_1: Cell, cell_2: Cell):
        """
        Calculate the area similarity of two cells
        Return value range: float[0, 1]
        """
        return min(cell_1.area, cell_2.area) / max(cell_1.area, cell_2.area)

    def calcEuclideanDistance(self, cell_1: Cell, cell_2: Cell):
        """
        Calculate the Euclidean distance between two cell center points
        Return value range: float(0,∞)
        The smaller the distance, the more similar it is, scaled to [0, 1) by the arctangent function
        The return value is the cosine value after normalization [0, π/2]. The smaller the return value, the lower the similarity
        """
        dist = cell_1.vector.EuclideanDistance(cell_2.vector)
        # return np.cos(np.arctan(dist) / (np.pi / 2))
        return dist

    def compareDicSimilar(self, cell_1: Cell, cell_2: Cell):
        """
        Compare dic similarity
        Return value range: float(0, 1)
        """
        dic1 = Vector(cell_1.feature.dic_intensity, cell_1.feature.dic_variance)
        dic2 = Vector(cell_2.feature.dic_intensity, cell_2.feature.dic_variance)
        if dic1 and dic2:
            return np.sin(self.normalize(dic1.cosSimilar(dic2)))
        else:
            return 0

    def compareMcySimilar(self, cell_1: Cell, cell_2: Cell):
        """
        Compare mcy similarity
        Return value range: float(0, 1)
        """
        mcy1 = Vector(cell_1.feature.mcy_intensity, cell_1.feature.mcy_variance)
        mcy2 = Vector(cell_2.feature.mcy_intensity, cell_2.feature.mcy_variance)
        if mcy1 and mcy2:
            return np.sin(self.normalize(mcy1.cosSimilar(mcy2)))
        else:
            return 0

    def compareShapeSimilar(self, cell_1: Cell, cell_2: Cell):
        """
        Calculate the contour similarity of two cells
        Return value range: float(0, 1)
        The smaller the score value, the greater the similarity, and the inversion operation can be performed.
        (The return value is the cosine value after normalization [0, π/2], no longer do this change, directly return
        the score.)
        """
        score = cv2.matchShapes(cell_1.contours, cell_2.contours, 1, 0.0)
        # return np.cos(self.normalize(score))
        return score

    def __str__(self):
        return f"Match object at {id(self)}"


class Matcher(object):
    """
    A matcher that realizes the association of front and back frame objects.
    """

    def __init__(self):
        self.matcher = Match()
        self.WEIGHT = {'IoU': 0.6, 'shape': 0.1, 'area': 0.3}

    def draw_bbox(self, bg1, bbox, track_id=None):
        if len(bg1.shape) > 2:
            im_rgb1 = bg1
        else:
            im_rgb1 = cv2.cvtColor(convert_dtype(bg1), cv2.COLOR_GRAY2RGB)
        cv2.rectangle(im_rgb1, (bbox[2], bbox[0]), (bbox[3], bbox[1]),
                      [255, 255, 0], 2)
        cv2.putText(im_rgb1, str(track_id), (bbox[3], bbox[1]), cv2.FONT_HERSHEY_COMPLEX,
                    1, (255, 255, 255), 1)
        return im_rgb1

    def predict_next_position(self, parent: Cell):
        """
        According to the speed, the possible position of the daughter cell is predicted, and the predicted daughter cell
        is used to participate in the matching. speed=0 is equivalent to not enabling prediction.
        """
        # new_cell = parent.move(parent.move_speed, 1)
        new_cell = parent
        return new_cell

    def _filter(self, child: Cell, cells: List[Cell]):
        """filter candidates"""

        # filtered_candidates = [cell for cell in cells if cell in child]
        filtered_candidates = []
        for cell in cells:
            if cell in child and cell.is_accurate_matched is False:
                if self.matcher.calcEuclideanDistance(cell, child) < child.r_long or \
                        self.matcher.calcIoU(cell, child) > 0:
                    filtered_candidates.append(cell)
        if not filtered_candidates:
            for cell in cells:
                if cell in child and cell.is_accurate_matched is False:
                    filtered_candidates.append(cell)

        return filtered_candidates

    def match_candidates(self, child: Cell, before_cell_list: List[Cell]):
        """match candidates"""
        return self._filter(child, before_cell_list)

    def calc_similar(self, parent, child_cell):
        """Calculate the similarity of two cells"""
        similar = [self.matcher.calcIoU(parent, child_cell),
                   self.matcher.calcAreaSimilar(parent, child_cell),
                   self.matcher.compareShapeSimilar(parent, child_cell),
                   self.matcher.calcEuclideanDistance(parent, child_cell),
                   self.matcher.calcCosDistance(parent, child_cell),
                   self.matcher.calcCosSimilar(parent, child_cell),
                   self.matcher.compareDicSimilar(parent, child_cell),
                   self.matcher.compareMcySimilar(parent, child_cell)]
        np.set_printoptions(precision=6, floatmode='fixed')
        return np.array(similar)

    @lru_cache(maxsize=None)
    def match_similar(self, cell_1: Cell, cell_2: Cell):
        """Calculate the similarity of two cells, do not include the image information."""
        similar = {'IoU': self.matcher.calcIoU(cell_1, cell_2),
                   'shape': self.matcher.compareShapeSimilar(cell_1, cell_2),
                   'area': self.matcher.calcAreaSimilar(cell_1, cell_2),
                   'distance': self.matcher.calcEuclideanDistance(cell_1, cell_2)
                   }
        return similar

    def match_duplicate_child(self, parent, unmatched_child_list):
        """
        Match multiple candidates. Calling this function means that there is more than one candidate, and this method
        calculates the matching degree of each candidate
        The return value is a dictionary in the form of {Cell: similar_dict}
        """
        matched = {}
        for i in unmatched_child_list:
            similar = self.match_similar(parent, i)
            if similar['IoU'] > 0:
                matched[i] = similar
        return matched

    def is_mitosis_start(self, pre_parent: Cell, last_leaves: List[Cell], area_size_t=1.5, iou_t=0.3):
        """
        Determine whether a cell enters the M cell_type. The basic criterion is that when a cell enters mitosis, its volume
        increases and the number of candidate regions increases. If the cell successfully enters the M cell_type, return a
        dict containing information about the last frame of G2 and the first frame of M. Otherwise, return False.
         """
        match_score = {}
        for i in last_leaves:
            if self.match_similar(pre_parent, i).get('IoU') >= iou_t:
                match_score[i] = self.match_similar(pre_parent, i)
        for child_cell in match_score:
            if (child_cell.area / pre_parent.area) >= area_size_t:
                return {'last_G2': pre_parent, 'first_M': child_cell}
        return False

    def get_similar_sister(self, parent: Cell, matched_cells_dict: dict, area_t=0.7, shape_t=0.03, area_size_t=1.3,
                           iou_t=0.1):
        """Find the two most similar cells among multiple candidates as the daughter cells."""
        cell_dict_keys = list(matched_cells_dict.keys())
        cell_dict_keys.sort(key=lambda cell: cell.area, reverse=True)

        for cell in cell_dict_keys:
            # if matched_cells_dict[cell].get('IoU') < iou_t:
            if matched_cells_dict[cell].get('IoU') == 0:
                cell_dict_keys.remove(cell)
        # if len(cell_dict_keys) > 2:
        #     if cell_dict_keys[0].area + cell_dict_keys[1].area > parent.area * area_size_t:
        #         cell_dict_keys.remove(cell_dict_keys[0])
        if len(cell_dict_keys) > 2:
            remove_dict_list = list(matched_cells_dict).sort(key=lambda x: x[1]['IoU'], reverse=True)[2:]
            for i in remove_dict_list:
                key = list(i.keys())[0]
                cell_dict_keys.remove(key)

        length = len(cell_dict_keys)
        match_result = {}
        for i in range(length - 1):
            cell_1 = cell_dict_keys.pop(0)
            for j in range(len(cell_dict_keys)):
                cell_2 = cell_dict_keys[j]
                score = self.match_similar(cell_1, cell_2)
                # if score.get('area') >= area_t and score.get('shape') <= shape_t:
                if score.get('area') >= area_t:
                    match_result[(cell_1, cell_2)] = score.get('area') + score.get('IoU')
        if match_result:
            max_score_cells = max(match_result, key=match_result.get)
            return max_score_cells
        else:
            raise MitosisError('cannot match the suitable daughter cells')

    def select_mitosis_cells(self, parent: Cell, candidates_child_list: List[Cell], area_t=0.5, shape_t=0.05,
                             area_size_t=1.3):
        """
        If cell division occurs, select two daughter cells. When calling this method, make sure that cell division is
        highly likely to occur. If the return value is cell division, then it is necessary to check whether the areas
        of the two daughter cells are normal. If the area of one daughter cell is too large, it is considered a FP.

        :return ([cell_1, cell2], 'match status')

        """

        matched_candidates = self.match_duplicate_child(parent, candidates_child_list)
        checked_candidates = {}
        for i in matched_candidates:
            if i.is_be_matched:
                if i.status.exit_mitosis_time < config.ENTER_DIVISION_THRESHOLD:
                    continue
            checked_candidates[i] = matched_candidates[i]
        matched_cells_dict = self.check_iou(checked_candidates)
        if not matched_cells_dict:
            raise MitosisError('not enough candidates')
        else:
            if len(matched_cells_dict) == 2:
                cells = list(matched_cells_dict.keys())
                if self.matcher.calcAreaSimilar(cells[0], cells[1]) > area_t:
                    return cells, 'ACCURATE'
                else:
                    return cells, 'INACCURATE'
            else:
                try:
                    max_two = self.get_similar_sister(parent, matched_cells_dict)  # 找到匹配结果中最大的两个值
                except MitosisError:
                    raise MitosisError("not enough candidates, after matched.")
                better_sisters = []
                for i in max_two:
                    better_sisters.append(i)
                if self.matcher.calcAreaSimilar(better_sisters[0],
                                                better_sisters[1]) > area_t and self.matcher.compareShapeSimilar(
                    better_sisters[0], better_sisters[1]) < shape_t:
                    return better_sisters, 'ACCURATE'
                else:
                    return better_sisters, 'INACCURATE'

    def select_single_child(self, score_dict):
        """
        For multiple IOU matching options, choose the one with a higher similarity. This is to distinguish overlapping
        cells rather than cell division. Note: the results of this method are not necessarily accurate and may result
        in mismatches due to cell crossing, which needs to be resolved in subsequent steps. In addition, if one cell is
        accurately matched, and the other cell has no match (i.e., not detected in the next frame during recognition),
        it should be filled as a predicted cell.

        """
        def calc_weight(candidate_score_dict):
            """Compute and dynamically update matching weights and matching results."""
            # for cell in candidate_score_dict:
            #     score_dict = candidate_score_dict[cell]
            # value =  score_dict['IoU'] * self.WEIGHT.get('IoU') + \
            #        score_dict['shape'] * self.WEIGHT.get('shape') + score_dict['area'] * self.WEIGHT.get('area')
            selected_cell = None
            max_iou = 0.0
            min_distance = 100
            threshold = 0.5
            iou_above_threshold = False

            for cell, score_dict in candidate_score_dict.items():
                iou = score_dict['IoU']
                distance = score_dict['distance']

                # If iou is greater than the threshold, update the maximum iou and the selected cell,
                # and set iou_above_threshold to True
                if iou > threshold:
                    iou_above_threshold = True
                    if distance < min_distance:
                        selected_cell = cell
                        min_distance = distance
                        max_iou = iou

                # If the iou is not greater than the threshold, but greater than the current maximum iou,
                # update the maximum iou and the selected cell
                elif iou > max_iou:
                    selected_cell = cell
                    max_iou = iou
                    min_distance = distance

            # If the iou is not greater than the threshold, but greater than the current maximum iou,
            # update the maximum iou and the selected cell
            if not iou_above_threshold:
                for cell, score_dict in candidate_score_dict.items():
                    iou = score_dict['IoU']
                    if iou > max_iou:
                        selected_cell = cell
                        max_iou = iou
            return selected_cell

        candidates = {}
        for cell in score_dict:
            if score_dict[cell].get('IoU') > 0.5:
                candidates[cell] = score_dict[cell]
        if not candidates:
            for cell in score_dict:
                if score_dict[cell].get('IoU') > 0.1:
                    candidates[cell] = score_dict[cell]
        if not candidates:
            for cell in score_dict:
                if score_dict[cell].get('IoU') > 0.0:
                    candidates[cell] = score_dict[cell]
        if not candidates:
            if not score_dict:
                return None
            for cell in score_dict:
                candidates[cell] = sum(score_dict[cell].values())
            best = max(candidates, key=candidates.get)
            return best
        else:
            best = calc_weight(candidates)
            return best

    def match_one(self, predict_child, candidates):
        if len(candidates) == 1:
            score = self.matcher.calcIoU(predict_child, candidates[0])
            if score > 0.5:
                return [(candidates[0], 'ACCURATE')]
            elif score > 0:
                return [(candidates[0], 'INACCURATE')]
            else:
                return [(candidates[0], 'INACCURATE')]
        else:
            return False

    def check_iou(self, similar_dict):
        """
        Check the IoU to provide a basis for determining cell division. If there are less than 2 matching options
        with IoU > 0, return False. Otherwise, return these two cells.
        """
        matched = {}
        for cell in similar_dict:
            score = similar_dict[cell]
            if score.get('IoU') > 0:
                matched[cell] = score
        if len(matched) < 2:
            return False
        else:
            return matched

    def _match(self, parent: Cell, filter_candidates_cells: List[Cell], cell_track_status: TreeStatus):
        """
        Compare the overall similarity of two cells.
        """
        predict_child = parent
        # filtered_candidates = self.match_candidates(predict_child, no_filter_candidates_cells)
        filtered_candidates = [cell for cell in filter_candidates_cells if cell.is_accurate_matched is False]
        if len(filtered_candidates) > 0:
            if not self.match_one(predict_child, filtered_candidates):  # 不只有一个选项
                if self.match_one(predict_child, filtered_candidates) is None:
                    return
                matched_candidates = self.match_duplicate_child(predict_child, filtered_candidates)
                if len(matched_candidates) > 1 and cell_track_status.exit_mitosis_time > 25:
                    if config.INCLUDE_CELL_CYCLE is False:
                        cell_track_status.enter_mitosis(parent.frame)
                if not cell_track_status.status.get('enter_mitosis'):
                    if self.select_single_child(matched_candidates) is None:
                        return
                    return {'matched_cell': [(self.select_single_child(matched_candidates), 'INACCURATE')],
                            'status': cell_track_status}
                else:
                    matched_result = []
                    try:
                        sisters, status = self.select_mitosis_cells(parent, filtered_candidates)  # 此时细胞一分为二
                        cell_track_status.exit_mitosis(parent.frame + 1)
                        for sister in sisters:
                            matched_result.append((sister, status))

                    except MitosisError as M:  # 细胞可能仍然处于M期，但是已经完成分开，或者只是被误判为M期
                        if self.select_single_child(matched_candidates):
                            matched_result.append((self.select_single_child(matched_candidates), 'INACCURATE'))
                    except ErrorMatchMitosis as M2:
                        # 细胞可能不均等分裂
                        if self.select_single_child(matched_candidates):
                            matched_result.append((self.select_single_child(matched_candidates), 'INACCURATE'))
                    finally:
                        if matched_result:
                            return {'matched_cell': matched_result, 'status': cell_track_status}
                        return None
            else:
                return {'matched_cell': self.match_one(predict_child, filtered_candidates), 'status': cell_track_status}

    def calc_sorted_value(self, parent: Cell, matched_cell):
        """Calculate the sorting value of a Cell object."""

        match_score = self.match_similar(parent, matched_cell)
        sort_value = match_score['IoU'] + 1 / (match_score['distance'] + 1e-5)
        return sort_value

    def add_child_node(self, tree, child_node: CellNode, parent_node: CellNode):
        try:
            tree.add_node(child_node, parent=parent_node)
            child_node.set_tree_status(tree.status)
        except treelib.exceptions.DuplicatedNodeIdError:
            pass

    def match_single_cell(self, tree: TrackingTree, current_frame: FeatureExtractor):
        """The implementation logic for tracking a single cell."""

        cells = current_frame.cells
        parents = tree.last_layer_cell
        m_counter = 5
        if len(parents) > 1:
            parent_dict = {parent: parent.frame for parent in parents}
            keys = [p for p in parents]
            if parent_dict[keys[0]] != parent_dict[keys[1]]:
                min_parent = min(parent_dict, key=parent_dict.get)
                parents.pop(min_parent)
        for parent in parents:
            # Two or more cells after cell division.
            tree.m_counter -= 1
            if not tree.m_counter:
                tree.m_counter = 5
                tree.status.reset_M_count()
            if parent.cell_type == 'M':
                tree.status.add_M_count()
            if tree.status.predict_M_len >= 3:
                tree.status.enter_mitosis(parent.frame - 3)
            predict_child = parent
            filtered_candidates = self.match_candidates(predict_child, cells)
            before_parent = tree.get_parent(parents[parent])
            if before_parent:
                if self.is_mitosis_start(before_parent.cell, [predict_child]):
                    tree.status.enter_mitosis(parent.frame)
            match_result = self._match(predict_child, filtered_candidates, tree.status)
            if match_result is not None:
                child_cells = match_result.get('matched_cell')
                for i in child_cells:
                    sort_value = self.calc_sorted_value(parent, i[0])
                    i[0].sort_value = sort_value
                parent_node = CellNode(parent)
            else:
                continue
            if len(child_cells) == 1:
                if child_cells[0][1] == 'ACCURATE':
                    child_cells[0][0].is_accurate_matched = True
                    child_node = CellNode(child_cells[0][0])
                else:
                    child_node = CellNode(child_cells[0][0])
                child_node.cell.set_branch_id(parent_node.cell.branch_id)
                child_node.cell.set_status(tree.status)
                child_node.cell.update_region(track_id=tree.track_id)
                child_node.cell.update_region(branch_id=parent_node.cell.branch_id)
                child_node.cell.set_match_status(child_cells[0][1])
                self.add_child_node(tree, child_node, parent_node)
            else:
                for cell in child_cells:
                    new_branch_id = tree.branch_id_distributor()
                    cell[0].set_branch_id(new_branch_id)
                    cell[0].update_region(track_id=tree.track_id)
                    cell[0].update_region(branch_id=new_branch_id)
                    cell[0].set_match_status(cell[1])
                    child_node = CellNode(cell[0])
                    if cell[1] == 'ACCURATE':
                        cell[0].is_accurate_matched = True
                    self.add_child_node(tree, child_node, parent_node)
        tree.status.add_exist_time()


class Tracker(object):
    """
    Tracker object, which is the controller of the tracking process. It reads images frame by frame,
    initializes and updates the TrackingTree, performs matching and assigns various IDs.
    """

    def __init__(self, annotation, mcy=None, dic=None):
        self.matcher = Matcher()
        self.fe_cache = deque(maxlen=5)
        self.trees: List[TrackingTree] = []
        self.mcy = mcy
        self.dic = dic
        self.annotation = annotation
        self._exist_tree_id = []
        self._available_id = 1
        self.init_flag = False
        self.feature_ext = feature_extract(mcy=self.mcy, dic=self.dic, jsonfile=self.annotation)
        self.tree_maps = {}
        self.init_tracking_tree(next(self.feature_ext)[0])
        self.nodes = set()
        self.count = 0
        self.parser_dict = None

    def id_distributor(self):
        if self._available_id not in self._exist_tree_id:
            self._exist_tree_id.append(self._available_id)
            self._available_id += 1
            return self._available_id - 1
        else:
            i = 1
            while True:
                if self._available_id + i not in self._exist_tree_id:
                    self._available_id += (i + 1)
                    return self._available_id + i
                i += 1

    def init_tracking_tree(self, fe: FeatureExtractor):
        """Initialize TrackingTree"""
        for i in fe.cells:
            tree = TrackingTree(track_id=self.id_distributor())
            i.sort_value = 1
            i.set_match_status('ACCURATE')
            i.set_track_id(tree.track_id, 1)
            i.set_branch_id(0)
            i.set_cell_id(str(i.track_id) + '-' + str(i.branch_id))
            i.update_region(track_id=tree.track_id)
            i.update_region(branch_id=0)
            node = CellNode(i)
            node.set_branch_id(0)
            node.set_track_id(tree.track_id)
            node.status = 'ACCURATE'
            i.set_match_status('ACCURATE')
            i.is_accurate_matched = True
            node.cell.set_status(TreeStatus(tree))
            tree.add_node(node)
            self.trees.append(tree)
            self.tree_maps[i] = tree
        self.init_flag = True
        self.fe_cache.append(fe)

    def draw_bbox(self, bg1, cell: Cell, track_id, branch_id=None, phase=None):
        bbox = cell.bbox
        if len(bg1.shape) > 2:
            im_rgb1 = bg1
        else:
            im_rgb1 = cv2.cvtColor(convert_dtype(bg1), cv2.COLOR_GRAY2RGB)
        cv2.rectangle(im_rgb1, (bbox[2], bbox[0]), (bbox[3], bbox[1]),
                      [0, 255, 0], 2, )

        def get_str():
            raw = str(track_id)
            if branch_id is not None:
                raw += '-'
                raw += str(branch_id)
            if phase is not None:
                raw += '-'
                raw += str(phase)
            return raw

        text = get_str()
        cv2.putText(im_rgb1, text, (bbox[3], bbox[1]), cv2.FONT_HERSHEY_TRIPLEX,
                    1, (255, 255, 255), 1)

        return im_rgb1

    @staticmethod
    def update_speed(parent: Cell, child: Cell, default: Vector = None):
        """Update cell movement speed"""
        if default:
            child.update_speed(Vector(0, 0))
        else:
            speed = (child.vector - parent.vector)
            child.update_speed(speed)

    def get_current_tree(self, parent_cell: Cell):
        """Get the TrackingTree where the current parent cell is located"""
        exist_trees = []  # Possible trees include the cells
        for tree in self.trees:
            if parent_cell in tree.last_layer_cell:
                exist_trees.append(tree)
        return exist_trees

    def add_node(self, child_node, parent_node, tree):
        if child_node not in tree:
            tree.add_node(child_node, parent=parent_node)
            child_node.cell.set_status(CellStatus(tree))
        else:
            raise NodeExistError(child_node)

    def track_near_frame(self, fe1: FeatureExtractor, fe2: FeatureExtractor):
        """match adjacent frames"""
        cells = sorted(fe1.cells, key=lambda cell: cell.sort_value, reverse=True)
        for parent in cells:
            trees = self.get_current_tree(parent)
            for tree in trees:
                self.matcher.match_single_cell(tree, fe2)

    def track_near_frame_mult_thread(self, fe1: FeatureExtractor, fe2: FeatureExtractor):
        """Match Adjacent Frames, Multithreaded Beta"""
        def work(__parent: Cell):
            trees = self.get_current_tree(__parent)
            for tree in trees:
                self.matcher.match_single_cell(tree, fe2)

        thread_pool_executor = ThreadPoolExecutor(max_workers=10, thread_name_prefix="track_")
        cells = sorted(fe1.cells, key=lambda cell: cell.sort_value, reverse=True)
        for parent in cells:
            thread_pool_executor.submit(work, parent)
        thread_pool_executor.shutdown(wait=True)

    def handle_duplicate_match(self, duplicate_match_cell):
        """Solve a cell is repeatedly matched by multiple cells"""
        child_node = CellNode(duplicate_match_cell)
        tmp = self.get_current_tree(duplicate_match_cell)
        parent0 = tmp[0].parent(child_node.nid)
        parent1 = tmp[1].parent(child_node.nid)
        tree_dict = {parent0: tmp[0], parent1: tmp[1]}
        sm0 = self.matcher.match_similar(duplicate_match_cell, parent0.cell)
        sm1 = self.matcher.match_similar(duplicate_match_cell, parent1.cell)
        match_score = {parent0: sm0['distance'],
                       parent1: sm1['distance']}
        error_parent = min(match_score)
        tree_dict[error_parent].remove_node(child_node.nid)
        return {error_parent: tree_dict[error_parent]}

    def rematch(self, fe1: FeatureExtractor, fe2: FeatureExtractor):
        """For the loss detection cells, re-match the upper and lower frames with the cache frame"""
        unmatched_list = [cell for cell in fe1.cells if cell.is_be_matched is False]
        if not unmatched_list:
            return
        for cell in unmatched_list:
            current_frame = cell.frame
            if cell.is_be_matched is False:
                handle_flag = False
                trees = self.trees
                wait_dict = {}
                wait_tree_map = {}
                for tree in trees:
                    last_layer_cells = [cell for cell in tree.last_layer_cell]
                    for last_layer_cell in last_layer_cells:
                        if 1 < current_frame - last_layer_cell.frame < 4:
                            match_result = self.matcher.match_similar(last_layer_cell, cell)
                            if match_result['IoU'] > 0 or match_result['distance'] < cell.d_long:
                                wait_dict[last_layer_cell] = match_result['IoU'] + 10 / (
                                        match_result['distance'] + 1e-5)
                                wait_tree_map[last_layer_cell] = tree
                                handle_flag = True
                    if (not handle_flag):
                        for last_layer_cell in last_layer_cells:
                            if config.GAP_WINDOW_LEN:
                                if 1 < current_frame - last_layer_cell.frame < config.GAP_WINDOW_LEN:
                                    match_result = self.matcher.match_similar(last_layer_cell, cell)
                                    if match_result['distance'] < 50:
                                        wait_dict[last_layer_cell] = match_result['IoU'] + 10 / (
                                                match_result['distance'] + 1e-5)
                                        wait_tree_map[last_layer_cell] = tree
                                        handle_flag = True
                            else:
                                if current_frame > last_layer_cell.frame:
                                    match_result = self.matcher.match_similar(last_layer_cell, cell)
                                    if match_result['distance'] < 50:
                                        wait_dict[last_layer_cell] = match_result['IoU'] + 10 / (
                                                match_result['distance'] + 1e-5)
                                        wait_tree_map[last_layer_cell] = tree
                                        handle_flag = True
                if wait_dict:
                    matched_cell = max(wait_dict, key=wait_dict.get)
                    tree = wait_tree_map[matched_cell]
                    child_node = CellNode(cell)
                    parent_node = CellNode(matched_cell)
                    tree.add_node(child_node, parent_node)
                    cell.is_accurate_matched = True
                    cell.is_be_matched = True
                    cell.set_status(tree.status)
                    cell.set_match_status('ACCURATE')
                    cell.set_track_id(tree.track_id, 1)
                    cell.set_branch_id(matched_cell.branch_id)
                    self.matcher.match_single_cell(tree, fe2)
                if not handle_flag:
                    tree = TrackingTree(track_id=self.id_distributor())
                    cell.set_track_id(tree.track_id, 1)
                    cell.set_branch_id(0)
                    cell.set_cell_id(str(cell.track_id) + '-' + str(cell.branch_id))
                    cell.update_region(track_id=tree.track_id)
                    cell.update_region(branch_id=0)
                    node = CellNode(cell)
                    node.set_branch_id(0)
                    node.set_track_id(tree.track_id)
                    node.status = 'ACCURATE'
                    tree.add_node(node)
                    self.trees.append(tree)
                    self.tree_maps[cell] = tree
                    cell.set_match_status('ACCURATE')
                    cell.set_status(CellStatus(tree))
                    cell.is_accurate_matched = True
                    cell.is_be_matched = True
                    self.matcher.match_single_cell(tree, fe2)

    def check_track(self, fe1: FeatureExtractor, fe2: FeatureExtractor, fe3: FeatureExtractor):
        """
        Check the track results to see if there are any wrong matches and omissions,
        and update the matching status at the same time
        """
        cells = sorted(fe2.cells, key=lambda cell: cell.sort_value, reverse=True)
        self.rematch(fe1, fe2)
        for cell in cells:
            tmp = self.get_current_tree(cell)
            if len(tmp) > 1:
                self.handle_duplicate_match(duplicate_match_cell=cell)

    def track(self, range=None, speed_filename=None):
        """Read image frames sequentially and start tracking"""
        global writer, speed_f
        index = 0
        if speed_filename:
            speed_f = open(speed_filename, 'w', newline='')
            writer = csv.writer(speed_f)
            writer.writerow(['Iteration', 'Speed (it/s)'])
        for fe_before, fe_current, fe_next in tqdm(self.feature_ext, total=range, desc='tracking process'):
            start_time = time.time()
            self.track_near_frame(fe_before, fe_current)
            # self.track_near_frame_mult_thread(fe_before, fe_current)
            self.fe_cache.append(fe_before)
            self.check_track(fe_before, fe_current, fe_next)
            del fe_before

            end_time = time.time()
            if speed_filename:
                writer.writerow([index, f"{1 / (end_time - start_time + 1e-10):.2f}"])
            if range:
                index += 1
                if index > range:
                    break
        if speed_filename:
            speed_f.close()

        __filter_trees = [tree for tree in self.trees if len(tree.nodes) > config.FILTER_THRESHOLD]
        del self.trees
        self.trees = __filter_trees

    def track_tree_to_json(self, filepath):
        if not os.path.exists(filepath):
            os.makedirs(filepath)
        for i in self.trees:
            jsf = os.path.join(filepath, f'tree-{i.track_id}.json')
            if os.path.exists(jsf):
                os.remove(jsf)
            i.save2file(jsf)

    def visualize_single_tree(self, tree, background_filename_list, save_dir, xrange=None):
        bg_fname = background_filename_list[:xrange + 2] if xrange else background_filename_list
        images = list(map(lambda x: cv2.imread(x, -1), bg_fname))
        images_dict = dict(zip(list(range(len(bg_fname))), images))
        for node in tree.expand_tree():
            frame = tree.nodes.get(node).cell.frame
            bbox = tree.nodes.get(node).cell.bbox
            img_bg = images_dict[frame]
            phase = tree.nodes.get(node).cell.cell_type
            images_dict[frame] = self.draw_bbox(img_bg, tree.nodes.get(node).cell, tree.track_id,
                                                tree.get_node(node).cell.branch_id, phase)
        for i in zip(bg_fname, list(images_dict.values())):
            fname = os.path.join(save_dir, os.path.basename(i[0]).replace('.tif', '.png'))
            cv2.imwrite(fname, i[1])

    def visualize_to_tif(self, background_mcy_image: str, output_tif_path, tree_list, xrange=None, single=False):
        """
        Visualize the tracking results, you can choose to save as a single tif file or multiple tif sequences
        :param background_mcy_image: background image for visualization
        :param output_tif_path: output file path
        :param tree_list: TrackingTree list to visualize
        :param xrange: visualization range
        :param single: Whether to choose to save as a single file
        """

        def adjust_gamma(__image, gamma=1.0):
            image = convert_dtype(__image)
            brighter_image = np.array(np.power((image / 255), 1 / gamma) * 255, dtype=np.uint8)
            return brighter_image

        tif = readTif(background_mcy_image)
        images_dict = {}
        index = 0

        if xrange is not None:
            for img, _ in tif:
                if index >= xrange:
                    break
                img = adjust_gamma(img, gamma=1.5)
                images_dict[index] = img
                index += 1
        else:
            try:
                for img, _ in tif:
                    images_dict[index] = img
                    index += 1
            except KeyError:
                pass
        for i in tree_list:
            for node in i.expand_tree():
                frame = i.nodes.get(node).cell.frame
                if xrange:
                    if frame > xrange:
                        continue
                # bbox = i.nodes.get(node).cell.bbox
                img_bg = images_dict.get(frame)
                if img_bg is not None:
                    images_dict[frame] = self.draw_bbox(img_bg, i.nodes.get(node).cell, i.track_id,
                                                        i.get_node(node).cell.branch_id,
                                                        phase=i.get_node(node).cell.cell_type)
        if not single:
            if not (os.path.exists(output_tif_path) and os.path.isdir(output_tif_path)):
                os.mkdir(output_tif_path)
            for i in tqdm(range(index), desc="save tracking visualization"):
                fname = os.path.join(output_tif_path, f'{os.path.basename(output_tif_path)[:-4]}-{i:0>4d}.tif')
                tifffile.imwrite(fname, images_dict[i])
        else:
            with tifffile.TiffWriter(output_tif_path) as tif:
                for i in tqdm(range(index)):
                    if i > 300:
                        warnings.warn(
                            "the image is to big to save, and the tifffile cannot save the size >4GB tifffile, "
                            "so this image will be cut down.")
                        break
                    tif.write(images_dict[i])


def get_cell_line_from_tree(tree: TrackingTree, dic_path: str, mcy_path: str, savepath):
    """
    Obtain a complete cell sequence from the track tree, including cell images, dic and mcy dual channels, and cycles,
    and the generated file name is named track_id-branch_id-frame-cell_type.tif"""
    if not os.path.exists(savepath):
        os.makedirs(savepath)
    save_mcy = os.path.join(savepath, 'mcy')
    save_dic = os.path.join(savepath, 'dic')
    if not os.path.exists(save_mcy):
        os.mkdir(save_mcy)
    if not os.path.exists(save_dic):
        os.mkdir(save_dic)
    mcy = tifffile.imread(mcy_path)
    dic = tifffile.imread(dic_path)
    for nid in tree.expand_tree():
        cell = tree.get_node(nid).cell
        y0, y1, x0, x1 = cell.bbox
        mcy_img = mcy[cell.frame][y0: y1, x0: x1]
        dic_img = dic[cell.frame][y0: y1, x0: x1]
        fname = str(tree.track_id) + '-' + str(cell.branch_id) + '-' + str(cell.frame) + '-' + str(
            cell.cell_type[0]) + '.tif'
        tifffile.imwrite(os.path.join(save_mcy, fname), convert_dtype(mcy_img))
        tifffile.imwrite(os.path.join(save_dic, fname), convert_dtype(dic_img))






# coding=utf-8

from __future__ import annotations

import glob
import logging
import os.path
import time

import numpy as np

from SCTrack import reclassification
from SCTrack.utils import mask_to_json


def start_track(fannotation: str | dict, fout, basename, track_range=None, fimage=None, fbf=None,
                export_visualization=True,
                track_to_json=True):
    """
     :param track_range: Track frame number range
     :param visualize_background_image: track background image
     :param basename:
     :param fannotation: segmentation output result, json file or dict
     :param fout: Tracking output folder path
     :param fimage: raw image path, can be empty
     :param fbf: Bright field image path, can be empty
     :param export_visualization: Whether to export the tracking visualization file, if yes, it will export a multi-frame tif file
     :param track_to_json: Whether to write the tracking result into fjson, if yes, a new json file will be generated
     :return: None
    """

    if type(fannotation) is str:
        if not fannotation.endswith('.json'):
            logging.info('convert mask to annotation file...')
            annotation = mask_to_json(fannotation, xrange=track_range)
        else:
            annotation = fannotation
    else:
        annotation = fannotation

    result_save_path = os.path.join(fout, 'tracking_output')
    if not os.path.exists(result_save_path):
        os.makedirs(result_save_path)
    reclassification.run(annotation=annotation, output_dir=result_save_path, track_range=track_range, dic=fbf,
                         mcy=fimage,
                         save_visualize=export_visualization, visualize_background_image=fimage,
                         track_to_json=track_to_json, basename=basename)

#
# def run():
#     """临时批处理函数，用完删除"""
#     # dirs = ['copy_of_1_xy01', 'copy_of_1_xy19', 'MCF10A_copy02', 'MCF10A_copy11', 'src06']
#     dirs = ['MCF10A_copy02_5min']
#     base = r'G:\paper\evaluate_data\5min'
#     for i in dirs:
#         image = glob.glob(rf"{base}\{i}\mcy*.tif")[0]
#
#         annotation = rf"{base}\{i}\5-result-GT.json"
#         print(image)
#         print(annotation)
#         output = rf"{base}\{i}"
#         start_track(annotation, output, os.path.splitext(os.path.basename(image))[0], None, image,
#                     export_visualization=False)
#         # cmd = fr"python main.py -p {image} -bf {image} -tp -ns -js {annotation}"
#         # print(cmd)
#
# def run_smooth():
#     image = r'G:\paper\evaluate_data\evaluate_smoothing\copy_of_1_xy19\mcy.tif'
#     ann = r"G:\paper\evaluate_data\evaluate_smoothing\copy_of_1_xy19\10%-GT.json"
#     out = r"G:\paper\evaluate_data\evaluate_smoothing\copy_of_1_xy19"
#     start_track(ann, out, os.path.splitext(os.path.basename(image))[0], None, image,
#                 export_visualization=False)


if __name__ == '__main__':
    # run()
    # run_smooth()
    # i = 8
    # annotation = rf"G:\paper\test\Data{i}\SEG.tif"
    # mcy_img = rf"G:\paper\test\Data{i}\01.tif"
    # start_track(annotation, rf"G:\paper\test\Data{i}", 'mcy', 1000,
    #             mcy_img)

    # image = r"G:\paper\evaluate_data\incorrect-data-test\normal-dataset\copy_of_1_xy19\mcy.tif"
    # annotation = r"G:\paper\evaluate_data\incorrect-data-test\normal-dataset\copy_of_1_xy19\result.json"
    #
    # outputdir = fr"G:\paper\evaluate_data\incorrect-data-test\normal-dataset\copy_of_1_xy19"
    # start_track(annotation, outputdir, os.path.splitext(os.path.basename(image))[0], None, image, export_visualization=True)

    # image = r"G:\paper\evaluate_data\evaluate_for_tracking\parameter_test_incorrect\!test_data\rpe19.tif"
    # annotation = r"G:\paper\evaluate_data\evaluate_for_tracking\parameter_test_incorrect\!test_data\result.json"
    #
    # outputdir = fr"G:\paper\evaluate_data\evaluate_for_tracking\parameter_test_incorrect\GAP_WINDOW_LEN\10"
    # start_track(annotation, outputdir, 'rpe19', None, image, export_visualization=False)
    # for i in [0]:
    #
    #     image = fr"G:\paper\evaluate_data\evaluate_noise_classification\group{i}\mcy.tif"
    #     annotation = fr"G:\paper\evaluate_data\evaluate_noise_classification\group{i}\GT.json"
    #
    #     outputdir = fr"G:\杂项\group{i}"
    #     start_track(annotation, outputdir, 'mcy', None, image, export_visualization=False)

    # break

    annotation = r"G:\paper\evaluate_data\evaluate_sctrack_param\MCF10A_copy02_5-result-GT.json"
    # annotation = r"G:\paper\evaluate_data\evaluate_sctrack_param\copy_of_1_xy01_5-result-GT.json"
    # mcy_img = r"G:\paper\evaluate_data\5min\copy_of_1_xy01_5min\mcy.tif"
    # dic_img = r"G:\raw data\demo\mcy.tif"
    start_track(annotation, r"G:\paper\evaluate_data\evaluate_sctrack_param\AVAILABLE_RANGE_COEFFICIENT\MCF10A_copy02", '0', None,
                None, export_visualization=False)



import json
import math
import os.path

import numpy as np
import cv2


class DetectionParser:
    def __init__(self, fname: str = None, py_data: dict = None):
        if fname:
            if fname.endswith('.json'):
                with open(fname) as jf:
                    self.__data = json.load(jf)
        elif py_data:
            if type(py_data) is dict:
                self.__data = py_data
        else:
            raise KeyError(f"fname or py_data need once")

    def get_frame_names(self):
        frames = list(self.__data.keys())
        return frames

    def get_regions_by_frame(self, frame_name):
        return self.__data[frame_name]['regions']

    def get_region_attr(self, regions):
        """
        regions: extract from json files, one frame data, it cna be get from function [get_regions_by_frame()]
        Returns: all cells bounding box and cell_type list,
        like [((x_min, x_max, y_min, y_max), cell_type), ..., ((x_min, x_max, y_min, y_max), cell_type)]

        """
        attrs = []
        for i in regions:
            all_x = i['shape_attributes']['all_points_x']
            all_y = i['shape_attributes']['all_points_y']
            phase = i['region_attributes']['cell_type']
            x_min = int(np.min(all_x))
            x_max = math.ceil(np.max(all_x))
            y_min = int(np.min(all_y))
            y_max = math.ceil(np.max(all_y))
            attrs.append([(x_min, x_max, y_min, y_max), phase])
        return attrs


def convert_dtype(__image: np.ndarray) -> np.ndarray:
    min_16bit = np.min(__image)
    max_16bit = np.max(__image)
    image_8bit = np.array(np.rint(255 * ((__image - min_16bit) / (max_16bit - min_16bit))), dtype=np.uint8)
    return image_8bit


def draw_bbox(image: np.ndarray, regions_bounding: list):
    if len(image.shape) > 2:
        im_rgb = image
    else:
        im_rgb = cv2.cvtColor(convert_dtype(image), cv2.COLOR_GRAY2RGB)
    _id = 0
    for i in regions_bounding:
        if i[1] == 'G1/G2':
            color = (0, 0, 255)
        elif i[1] == 'S':
            color = (0, 255, 0)
        else:
            color = (255, 0, 0)
        cv2.rectangle(im_rgb, (i[0][0], i[0][2]), (i[0][1], i[0][3]), color, 2)
        cv2.putText(im_rgb, str(_id), (i[0][0], i[0][2]), cv2.FONT_HERSHEY_COMPLEX, 1, (255, 255, 255), 1)
        _id += 1

    return im_rgb


def json2box(shapes: dict):
    pass


if __name__ == '__main__':
    pass
    # from tqdm import tqdm
    #
    # file = r"G:\20x_dataset\copy_of_xy_01\copy_of_1_xy01.json"
    # image = r'G:\20x_dataset\copy_of_xy_01\tif\mcy\copy_of_1_xy01-0000.tif'
    # base = r'G:\20x_dataset\copy_of_xy_01\tif\mcy'
    # dp = DetectionParser(fname=file)
    # dp.get_frame_names()
    # # rb = dp.get_region_attr(dp.get_regions_by_frame('copy_of_1_xy01-0000.png'))
    # im = cv2.imread(image, -1)
    # bb_list = []
    # videoWriter = cv2.VideoWriter(r'G:\20x_dataset\copy_of_xy_01\development-dir\result-2-multi-bbox.mp4',
    #                               cv2.VideoWriter_fourcc('M', 'P', '4', 'V'),
    #                               5, (2048, 2048), True)
    # for i in tqdm(range(len(dp.get_frame_names()))):
    #     fname = dp.get_frame_names()
    #     img = cv2.imread(os.path.join(base, fname[i].replace('.png', '.tif')), -1)
    #     if i == 0:
    #         rb = dp.get_region_attr(dp.get_regions_by_frame(fname[i]))
    #         rb_before = rb
    #     else:
    #         rb = dp.get_region_attr(dp.get_regions_by_frame(fname[i]))
    #         rb_before = dp.get_region_attr(dp.get_regions_by_frame(fname[i - 1]))
    #     tmp = draw_bbox(img, rb_before)
    #     ret = draw_bbox(tmp, rb)
    #     videoWriter.write(ret)
    # videoWriter.release()



#! /usr/bin/python3
# -*- coding: utf-8 -*-
# @FileName: main.py
# @Author: Li Chengxin 
# @Time: 2023/7/2 19:51


import argparse
import logging
import os
import sys
import time
import imagesize
from SCTrack import track, config


def main():
    logging.basicConfig(format="%(asctime)s %(levelname)s: %(message)s", datefmt="%Y-%m-%d %H:%M:%S",
                        level=logging.INFO)

    parser = argparse.ArgumentParser(description="", add_help=False)
    help_content = """
        Welcome to use SC-Track!
        using this script to auto tracking the single cell images and identify each cell's type.\n
        usage:
            sctrack -i <image filepath> -a <annotation filepath, json annotation or segmentation mask filepath>  
            -bf[optional] <bright field image filepath> -o [optional] <output result filepath> 
            -t [optional]
    """

    parser.add_argument("-h", "--help", action="help", help=help_content)
    parser.add_argument('-i', "--image", default=False, help="input image filepath for nuclear")
    parser.add_argument('-o', "--output", default=False, help='output json file path')
    parser.add_argument('-bf', "--bf", default=False, help='input image filepath of bright field')
    parser.add_argument('-ot', "--ot", default=False, help='tracking output result saved dir')
    parser.add_argument('-a', "--annotation", default=False, help='annotation file path, json file or tiff mask file.')
    parser.add_argument('-r', "--range", default=False,
                        help='tracking frame range, default is None, means tracking the whole timelapse')

    args = parser.parse_args()

    if len(sys.argv) < 2:
        parser.print_help()
        sys.exit(0)

    if args.image is False:
        logging.warning("Raw image not be given, so that the visualization result will not be support!")
        visualization = False
    else:
        image = args.image
        visualization = config.EXPORT_TRACKING_VISUALIZATION
        image_width, image_height = imagesize.get(image)
        bf = args.bf

    if args.annotation is False:
        logging.error(
            'You must provide segmentation annotation file for tracking, support json format or mask tif format.')
        sys.exit(-1)
    else:
        annotation = args.annotation

    if args.output is False:
        output = os.path.join(os.path.dirname(annotation), 'output.json')
        logging.warning(f"-o  not provided, using the default output file name: {output}")
    else:
        if not args.output.endswith('.json'):
            output = os.path.join(os.path.dirname(args.image), 'output.json')
        else:
            output = args.output

    start_time = time.time()

    if args.ot:
        track_output = args.ot
    else:
        track_output = os.path.dirname(output)

    if args.range is False:
        xrange = None
    else:
        try:
            xrange = int(args.range)
        except ValueError:
            logging.error(f'param <-r/--range >={args.range}, the value must be int!')
            sys.exit(-1)
    logging.info(f"Tracking result will saved to {track_output}\\tracking_output.")
    logging.info('start tracking ...')
    track.start_track(fannotation=annotation, fimage=args.image, fbf=None, fout=track_output, track_range=xrange,
                      export_visualization=visualization,
                      basename=os.path.splitext(os.path.basename(args.annotation))[0])

    end_time = time.time()

    logging.info(f'tracking cost time: {end_time - start_time:.4f}')


if __name__ == '__main__':
    main()



import logging


class InsertError(Exception):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return f"Insert Error: {self.value}!"

    def __repr__(self):
        return self.__str__()


class MitosisError(Exception):
    def __init__(self, message=None):
        self.message = message

    def __str__(self):
        return f"cannot match sister cells! {self.message if self.message else ''}!"

    def __repr__(self):
        return self.__str__()


class ErrorMatchMitosis(Exception):
    def __init__(self, message=None):
        self.message = message

    def __str__(self):
        return f"Unreasonable cell size, should not happen cell division! {self.message if self.message else ''}!"

    def __repr__(self):
        return self.__str__()


class NodeExistError(Exception):
    def __init__(self, node, message=None):
        self.message = message
        self.node = node

    def __str__(self):
        return f"cannot add node: {self.node}, already exists! {self.message if self.message else ''}!"

    def __repr__(self):
        return self.__str__()


class MatchFailed(Exception):
    def __init__(self, message=None):
        self.message = message

    def __str__(self):
        return f"match failed! {self.message if self.message else ''}!"

    def __repr__(self):
        return self.__str__()


class StatusError(Exception):
    def __init__(self, message=None):
        self.message = message

    def __str__(self):
        if self.message:
            return self.message
        else:
            return "when set the status, some error occurred..."

    def __repr__(self):
        return self.__str__()



from __future__ import annotations

import math
import warnings
from enum import Enum
from shapely.geometry import Polygon
import numpy as np
from functools import lru_cache
from SCTrack.utils import convert_dtype
from SCTrack import config


def NoneTypeFilter(func):
    def _filter(self, *args, **kwargs):
        ret = func(self, *args, **kwargs)
        cells = []
        for i in ret:
            if i.mcy.size != 0:
                cells.append(i)
        return cells

    return _filter


def warningFilter(func):
    warnings.filterwarnings("ignore")

    def _warn_filter(*args, **kwargs):
        return func(*args, **kwargs)

    return _warn_filter


class MatchStatus(Enum):
    """
    Matching status enumeration object,
    including matched, unmatched, and missing matches,
    is the status value of TrackingTree.
    """

    Matched = 0
    Unmatched = 1
    LossMatch = 2


class TreeStatus(object):
    """
    Record the state of the tracked cells, including cell cell_type status, division status, and matching status
    Phase: whether Mitosis has entered, and which frame has entered Mitosis
    Division: whether there is Mitosis
    Matching situation: Is there any loss of matching in this cell
    """

    # TreeStatus all Status properties
    __status_types = ['enter_mitosis', 'enter_mitosis_frame', 'division_event_happen',
                      'division_count', 'exit_mitosis', 'exit_mitosis_frame']

    _instances = {}

    def __new__(cls, *args, **kwargs):
        key = str(args) + str(kwargs)
        if key not in cls._instances:
            cls._instances[key] = super().__new__(cls)
            cls._instances[key].__tracking_tree = None
            cls._instances[key].__init_flag = False
            cls._instances[key].enter_mitosis_threshold = config.ENTER_DIVISION_THRESHOLD

            # The cell division window stage, with an initial value of 20, will exit division matching when no cell
            # division is matched or cell division is completed during this window period. After entering the split
            # window period, this value will decrease every frame forward.
            cls._instances[key].division_windows_len = 20

            # Starting from the completion of splitting and exiting mitosis, counting, no further entry into mitosis
            # is allowed within 10 frames, that is, when __exit_mitosis_time < 10, self.__ enter_ mitosis cannot be true
            cls._instances[key].__exit_mitosis_time = cls._instances[key].enter_mitosis_threshold

        return cls._instances[key]

    def __init__(self, tree: 'TrackingTree | None'):
        if not self.__init_flag:
            self.__tracking_tree = tree
            self.__enter_mitosis: bool = False
            self.__enter_mitosis_frame: int | None = None
            self.__division_event_happen: bool = False  # True indicates that the cell has at least one M    itosis
            self.__division_count: int = 0
            self.__exit_mitosis: bool = False
            self.__exit_mitosis_frame: int | None = None
            self.__match_status = MatchStatus.Unmatched

            # This value records the number of predicted M cell_type. If > 3, it is considered to have entered M period,
            # and at this point, enter needs to be called externally_ Mitosis()
            self.__predict_M_count = 0
            self.__init_flag = True

    @property
    def status(self):
        return dict(zip(TreeStatus.__status_types,
                        (self.__enter_mitosis, self.__enter_mitosis_frame, self.__division_event_happen,
                         self.__division_count, self.__exit_mitosis, self.__exit_mitosis_frame)))

    def get_status(self, status_type: str):
        """
        :param status_type: __status_types member
        :return: member status
        """
        return self.status.get(status_type)

    def check_division_window(self):
        """
        Check if the cell is in the division window stage
        """
        if self.division_windows_len > 0:
            return True
        return False

    def reset_division_window(self):
        """
        reset the division_windows_len
        """
        self.division_windows_len = 20

    def sub_division_window(self):
        """
        When the cell is in the division window stage,  continuously decreases the division_windows_len
        """
        if self.division_windows_len > 0:
            self.division_windows_len -= 1

    def is_in_division_window(self):
        if self.division_windows_len > 0:
            return True
        return False

    def enter_mitosis(self, frame):
        if self.__exit_mitosis_time >= self.enter_mitosis_threshold:
            self.__enter_mitosis = True
            self.__exit_mitosis = False
            self.__exit_mitosis_time = 0
            self.__enter_mitosis_frame = frame
            self.__exit_mitosis_frame = None
            self.reset_division_window()

    def exit_mitosis(self, frame):
        self.__enter_mitosis = False
        self.__exit_mitosis = True
        self.__exit_mitosis_frame = frame
        self.__division_event_happen = True
        self.__division_count += 1

    def set_matched_status(self, value):
        if value == 'matched':
            self.__match_status = MatchStatus.Matched
        elif value == 'loss':
            self.__match_status = MatchStatus.LossMatch
        else:
            pass

    def add_M_count(self):
        self.__predict_M_count += 1

    def reset_M_count(self):
        self.__predict_M_count = 0

    @property
    def predict_M_len(self):
        return self.__predict_M_count

    def add_exist_time(self):
        if self.__exit_mitosis:
            self.__exit_mitosis_time += 1

    @property
    def is_mitosis_enter(self):
        return self.__enter_mitosis

    @property
    def exit_mitosis_time(self):
        return self.__exit_mitosis_time

    def __str__(self):
        return str(self.status)

    def __repr__(self):
        return self.__str__()


class CellStatus(TreeStatus):
    """
    Record the status of cells and exclude them from other matching candidates if they have participated in precise matching.
    If the cell has undergone Mitosis in a short time, it will not participate in Mitosis matching.
    """
    pass


class SingleInstance(object):
    """Singleton pattern base class. If the parameters are the same, only one instance object will be instantiated"""
    _instances = {}
    init_flag = False

    def __new__(cls, *args, **kwargs):
        key = str(args) + str(kwargs)
        if key not in cls._instances:
            cls._instances[key] = super().__new__(cls)
            SingleInstance.__init__(SingleInstance._instances[key], *args, **kwargs)
        return cls._instances[key]

    def __init__(self, *args, **kwargs):
        pass


class Rectangle(object):
    """
    Rectangular class , used to record the bounding box and available range of cells
    """

    def __init__(self, x_min, x_max, y_min, y_max):
        self.x_min = x_min
        self.x_max = x_max
        self.y_min = y_min
        self.y_max = y_max

    @property
    def area(self):
        """
        Return the Rectangle area,
        """
        return abs((self.x_max - self.x_min) * (self.y_max - self.y_min))

    def _intersectX(self, other):
        """
        Determine whether two rectangles intersect on the X-axis
        """
        if max(self.x_min, other.x_min) - min(self.x_max, other.x_max) >= 0:
            return False
        else:
            return True

    def _intersectY(self, other):
        """Determine whether two rectangles intersect on the Y-axis"""
        if max(self.y_min, other.y_min) - min(self.y_max, other.y_max) >= 0:
            return False
        else:
            return True

    def _include(self, other, self_max, other_max, self_min, other_min):
        """
        Determine if it contains, incoming x value represents in the x-axis direction, incoming y value represents in the y-axis direction
        """
        if self_min >= other_min:
            if self_max <= other_max:
                flag = other
                return True, flag  # indicate longer instance
            else:
                return False
        else:
            if self_min <= other_min:
                if self_max >= other_max:
                    flag = self
                    return True, flag
                else:
                    return False

    def _includeX(self, other):
        """
        Determine if there is an inclusion relationship between the X-axis of two rectangles
        """
        return self._include(other, self.x_max, other.x_max, self.x_min, other.x_min)

    def _includeY(self, other):
        """
        Determine if there is an inclusion relationship between the Y-axis of two rectangles
        """
        return self._include(other, self.y_max, other.y_max, self.y_min, other.y_min)

    def isIntersect(self, other):
        """
        Determine whether two rectangles intersect
        """
        if self._intersectX(other) and self._intersectY(other):
            return True
        else:
            return False

    def isInclude(self, other):
        """
        Determine whether two rectangles contain
        """
        if not self._includeX(other):
            return False
        else:
            if not self._includeY(other):
                return False
            else:
                if self._includeX(other)[1] != self._includeY(other)[1]:
                    return False
                else:
                    return True

    def draw(self, background=None, isShow=False, color=(255, 0, 0)):
        """
        draw the rectangle
        """
        import cv2
        if background is not None:
            _background = background
        else:
            _background = np.zeros(shape=(100, 100))

        if len(_background.shape) == 2:
            _background = cv2.cvtColor(convert_dtype(_background), cv2.COLOR_GRAY2RGB)
            # (bbox[2], bbox[0]), (bbox[3], bbox[1])
        cv2.rectangle(_background, (self.y_min, self.x_min), (self.y_max, self.x_max), color, 2)
        if isShow:
            cv2.imshow('rec', _background)
            cv2.waitKey()
        return _background


class Vector(np.ndarray):
    """
    2D plane vectoer class
    """

    def __new__(cls, x=None, y=None, shape=(2,), dtype=float, buffer=None, offset=0,
                strides=None, order=None):
        obj = super().__new__(cls, shape, dtype,
                              buffer, offset, strides, order)
        if x is None and y is None:
            obj.xy = None
            obj.x = x
            obj.y = y
        elif x is None and y:
            obj.xy = [0, y]
            obj.x = 0
            obj.y = y
        elif x and y is None:
            obj.xy = [x, 0]
            obj.x = x
            obj.y = 0
        else:
            obj.xy = [x, y]
            obj.x = x
            obj.y = y
        return obj

    def __array_finalize__(self, obj):
        if obj is None:
            return
        self.info = getattr(obj, 'xy', None)
        self.x = getattr(obj, 'x', None)
        self.y = getattr(obj, 'y', None)

    @property
    def module(self):
        """
        Return the modulus of vector
        """
        if self.xy is None:
            return 0
        return np.linalg.norm([self.x, self.y])

    @property
    def cos(self):
        """
        Return the cosine value of vector
        """
        if not self.xy:
            return None
        return self.x / self.module

    def cosSimilar(self, other):
        """
        Compare the cosine similarity of two vectors, with a range of values [-1, 1]
        """
        if self.xy is None:
            if other.xy is None:
                return None
            elif other == Vector(0, 0):
                return 0
            else:
                return other.cos
        if self == other:
            return 1
        if other == Vector(0, 0):
            return 0
        if self == Vector(0, 0):
            if other.yx is None:
                return 0
            elif other == Vector(0, 0):
                return 1
            else:
                return other.cos
        return np.dot(self.xy, other.xy) / (self.module * other.module)

    def cosDistance(self, other):
        """
        Cosine distance, numerically equal to 1 minus cosine similarity, range of values [0,2]
        """
        return 1 - self.cosSimilar(other)

    def EuclideanDistance(self, other):
        """
        Return the Euclidean distance between two vectors
        """
        return np.sqrt((abs(self.x - other.x) ** 2 + abs(self.y - other.y) ** 2))

    def __len__(self):
        if self.xy is None:
            return 0
        return len(self.xy)

    def __str__(self):
        if self.xy is None:
            return "None Vector"
        return str(self.xy)

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        return self.x == other.x and self.y == other.y

    def __lt__(self, other):
        return self.module < other.module

    def __bool__(self):
        if not self.xy or self == Vector(0, 0):
            return False
        return True

    def __add__(self, other):
        return Vector(self.x + other.x, self.y + other.y)

    def __sub__(self, other):
        return Vector(self.x - other.x, self.y - other.y)

    def __neg__(self):
        return Vector(-self.x, -self.y)

    def __mul__(self, other):
        if hasattr(other, '__float__'):
            return Vector(self.x * other, self.y * other)
        elif isinstance(other, Vector):
            return self.x * other.x + self.y * other.y
        else:
            raise TypeError(f'{type(other)} are not support the Multiplication operation with type Vector!')


class Cell(object):
    """
    Define the cell class, which is the core class of SC-Track. The smallest unit of all operations is the Cell instance,
    and the Cell is implemented as a conditional singleton mode: that is, different objects are generated according to
    the parameters passed in. If the parameters passed in are the same , only one object is instantiated. When defining
    the Cell object, Cell class can be guaranteed that there is only one Cell instance for the same cell in one frame,
    and differentCell instances are generated in different frames.
    If you want to instantiate a Cell object, at least its position information needs to be passed in, that is, all the
    xy coordinate points that constitute the outline of the cell.
    """
    _instances = {}

    def __new__(cls, *args, **kwargs):
        key = str(args) + str(kwargs)
        if key not in cls._instances:
            cls._instances[key] = super(Cell, cls).__new__(cls)
            cls._instances[key].__feature = None
            cls._instances[key].sort_value = 0
            cls._instances[key].__feature_flag = False
            cls._instances[key].__track_id = None
            cls._instances[key].__branch_id = None
            cls._instances[key].__inaccurate_track_id = None
            cls._instances[key].__is_track_id_changeable = True
            cls._instances[key].mitosis_start_flag = False
            cls._instances[key].__region = None
            cls._instances[key].__status = None
            cls._instances[key].__is_accurate_matched = False
            cls._instances[key].__match_status = False  # 匹配状态，如果参与匹配则设置为匹配状态，从未匹配则设置为False
        return cls._instances[key]

    def __init__(self, position=None, mcy=None, dic=None, cell_type=None, frame_index=None, flag=None):
        # if  Cell.init_flag is False:
        self.position = position  # [(x1, x2 ... xn), (y1, y2 ... yn)]
        self.mcy = mcy
        self.dic = dic
        self.cell_type = cell_type
        self.__id = None
        self.frame = frame_index
        self.__parent = None  # If a cell divides, record the __id of the cell's parent
        self.__move_speed = Vector(0, 0)
        self.polygon = Polygon([xy for xy in zip(*self.position)])

        if flag is None:
            self.flag = 'cell'
        else:
            self.flag = 'gap'
        Cell.init_flag = True
        # else:
        #     return

    def change_mitosis_flag(self, flag: bool):
        """
        When the cell enters mitosis for the first time, self.mitosis_start_flag is set to True, and when the cell
         completes division, it is reset to false.
        """
        self.mitosis_start_flag = flag

    @property
    @lru_cache(maxsize=None)
    def contours(self):
        """
        Convert the list of xy coordinate points into contour points list
        :return: if successful, return contours, else return None
        """
        points = []
        if self.position:
            for j in range(len(self.position[0])):
                x = int(self.position[0][j])
                y = int(self.position[1][j])
                points.append((x, y))
            contours = np.array(points)
            return contours
        else:
            return None

    @property
    def move_speed(self) -> Vector:
        """
        :return:  moving speed of the cell, the Vector object type
        """
        return self.__move_speed

    def update_speed(self, speed: Vector):
        """
        :param speed: new speed of the cell, Vector instance.
        :return: None
        """
        self.__move_speed = speed

    @staticmethod
    def polygon_centroid(vertex_coordinates):
        """
        Calculate the physical center of gravity of the cell counters.
        :param vertex_coordinates: list of cell outline coordinate points, the format is [[x1, x2,...xn], [y1, y2,..yn]]
        :return: physical center of gravity
        """
        x_coords = vertex_coordinates[0]
        y_coords = vertex_coordinates[1]
        n = len(x_coords)
        area = 0.0
        centroid_x = 0.0
        centroid_y = 0.0
        for i in range(n):
            j = (i + 1) % n
            cross_product = x_coords[i] * y_coords[j] - x_coords[j] * y_coords[i]
            area += cross_product
            centroid_x += (x_coords[i] + x_coords[j]) * cross_product
            centroid_y += (y_coords[i] + y_coords[j]) * cross_product
        area /= 2.0
        centroid_x /= 6.0 * area
        centroid_y /= 6.0 * area
        return centroid_x, centroid_y

    @property
    @lru_cache(maxsize=None)
    def center(self):
        """
        :return: cell physical center
        """
        # return np.mean(self.position[0]), np.mean(self.position[1])
        return self.polygon_centroid(self.position)

    @property
    @lru_cache(maxsize=None)
    def available_range(self) -> Rectangle:
        """
        Define the matchable range of the two frames before and after, the default is twice the horizontal and vertical
        coordinates of the cell
        :return: Rectangle instance
        """
        mult = config.AVAILABLE_RANGE_COEFFICIENT

        x_len = self.bbox[3] - self.bbox[2]
        y_len = self.bbox[1] - self.bbox[0]
        x_min_expand = self.bbox[2] - mult * x_len
        x_max_expand = self.bbox[3] + mult * x_len
        y_min_expand = self.bbox[0] - mult * y_len
        y_max_expand = self.bbox[1] + mult * y_len
        return Rectangle(y_min_expand, y_max_expand, x_min_expand, x_max_expand)

    @property
    def r_long(self):
        """
        :return: Cell bounding box long side radius
        """
        return max((self.bbox[3] - self.bbox[2]) / 2, (self.bbox[1] - self.bbox[0]) / 2)

    @property
    def r_short(self):
        """
        :return: Cell bounding box short side radius
        """
        return min((self.bbox[3] - self.bbox[2]) / 2, (self.bbox[1] - self.bbox[0]) / 2)

    @property
    def d_long(self):
        """
        :return: Long side diameter of cell bounding box
        """
        return 2 * self.r_long

    @property
    def d_short(self):
        """
        :return: Short side diameter of cell bounding box
        """
        return 2 * self.r_short

    @staticmethod
    @lru_cache(maxsize=None)
    def polygon_area(x, y):
        """
        Calculate cell area

        :param x: A list of x-coordinates of all points of the cell counters
        :param y: A list of y-coordinates of all points of the cell counters
        :return: Area of cell
        """
        return 0.5 * np.abs(np.dot(x, np.roll(y, 1)) - np.dot(y, np.roll(x, 1)))

    @property
    @lru_cache(maxsize=None)
    def vector(self) -> Vector:
        """
        Returns the cell center point vector, starting from the upper left corner of the image as the origin.
        :return: Vector instance
        """
        return Vector(*self.center)

    @property
    def area(self):
        """
        :return: Area of a cell
        """
        return self.polygon_area(tuple(self.position[0]), tuple(self.position[1]))

    def set_region(self, region):
        self.__region = region

    def set_status(self, status: TreeStatus | CellStatus):
        self.__status = status

    @property
    def status(self):
        return self.__status

    @property
    def is_accurate_matched(self):
        return self.__is_accurate_matched

    @is_accurate_matched.setter
    def is_accurate_matched(self, value: bool):
        self.__is_accurate_matched = value

    @property
    def region(self):
        return self.__region

    @property
    @lru_cache(maxsize=None)
    def bbox(self):
        """bounding box coordinates"""
        x0 = math.floor(np.min(self.position[0])) if math.floor(np.min(self.position[0])) > 0 else 0
        x1 = math.ceil(np.max(self.position[0]))
        y0 = math.floor(np.min(self.position[1])) if math.floor(np.min(self.position[1])) > 0 else 0
        y1 = math.ceil(np.max(self.position[1]))
        return y0, y1, x0, x1

    def move(self, speed: Vector, time: int = 1):
        """
        :param speed: move speed， Vector object instance
        :param time: move time，frame
        :return: New Cell instance after moving
        """
        new_position = [tuple([i + speed.x * time for i in self.position[0]]),
                        tuple([j + speed.y * time for j in self.position[1]])]
        new_cell = Cell(position=new_position, mcy=self.mcy, dic=self.dic, cell_type=self.cell_type, frame_index=self.frame)
        new_cell.set_track_id(self.__track_id, 0)
        return new_cell

    def set_feature(self, feature):
        """set Feature object for cell"""
        self.__feature = feature
        self.__feature_flag = True

    @property
    def feature(self):
        if self.__feature_flag:
            return self.__feature
        else:
            raise ValueError("No available feature! ")

    def set_track_id(self, __track_id, status: 0 | 1):
        """Set the track_id for the cell"""
        if self.__is_track_id_changeable:
            if status == 1:
                self.__track_id = __track_id
                self.__is_track_id_changeable = False
            elif status == 0:
                self.__inaccurate_track_id = __track_id
            else:
                raise ValueError(f'status {status} is invalid!')
        else:
            # warnings.warn('cannot change the accurate track_id')
            pass

    def set_match_status(self, status: bool | str):
        self.__match_status = status

    @property
    def is_be_matched(self):
        """I
        f participated in the match, return match status, otherwise, False
        """
        return self.__match_status

    @is_be_matched.setter
    def is_be_matched(self, value):
        self.__match_status = value

    def set_parent_id(self, __parent_id):
        self.__parent = __parent_id

    def set_cell_id(self, cell_id):
        self.__id = cell_id

    def set_branch_id(self, branch_id):
        self.__branch_id = branch_id

    def update_region(self, **kwargs):
        """update annotation region information ,add the tracking results."""
        new_region = self.region
        if new_region:
            if 'branch_id' in kwargs:
                new_region['region_attributes']['branch_id'] = kwargs['branch_id']
            if 'track_id' in kwargs:
                new_region['region_attributes']['track_id'] = kwargs['track_id']
            self.set_region(new_region)

    @property
    def branch_id(self):
        return self.__branch_id

    @property
    def cell_id(self):
        """Cell id, parent cell and daughter cell have different value"""
        return self.__id

    @property
    def parent(self):
        return self.__parent

    @property
    def track_id(self):
        return self.__track_id

    def draw(self, background=None, isShow=False, color=(255, 0, 0)):
        import cv2
        if background is not None:
            _background = background
        else:
            _background = np.ones(shape=(2048, 2048, 3), dtype=np.uint8)

        if len(_background.shape) == 2:
            _background = cv2.cvtColor(convert_dtype(_background), cv2.COLOR_GRAY2RGB)
        # cv2.drawContours(_background, self.contours, -1, color, 3)
        cv2.rectangle(_background, (self.bbox[2], self.bbox[0]), (self.bbox[3], self.bbox[1]), color, 5)
        if isShow:
            cv2.imshow('rec', _background)
            cv2.resizeWindow('rec', 500, 500)
            cv2.waitKey()
        return _background

    def __contains__(self, item):
        return True if self.available_range.isIntersect(Rectangle(*item.bbox)) else False

    def __str__(self):
        if self.position:
            return f" Cell at ({self.center[0]: .2f},{self.center[1]: .2f}), frame {self.frame}, {self.cell_type}, branch {self.__branch_id}"
        else:
            return "Object Cell"

    def __repr__(self):
        return self.__str__()

    def __lt__(self, other):
        if self.position and other.position:
            self_module = np.linalg.norm([np.mean(self.position[0]), np.mean(self.position[1])])
            other_module = np.linalg.norm([np.mean(other.position[0]), np.mean(other.position[1])])
            return self_module < other_module
        else:
            raise ValueError("exist None object of ")

    def __eq__(self, other):
        return self.position == other.position and self.frame == other.frame

    def __hash__(self):
        return id(self)



import logging
import os.path
from collections import Counter
from copy import deepcopy
from typing import List
import json

import cv2
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import tifffile
from imagesize import imagesize

from SCTrack.base import Cell
from SCTrack.tracker import Tracker, CellNode, TrackingTree
from SCTrack.config import RAW_INPUT_IMAGE_SIZE
from SCTrack.config import N_CLASS, CLASS_NAME
from SCTrack import config


class TreeParser(object):
    """parse TrackingTree"""

    def __init__(self, track_tree: TrackingTree):
        self.tree = track_tree

        self.lineage_dict = {}
        # {cells: [children_list], branch_id: ''， 'G1_start': '', 'S_start': '', 'G2_start': '', 'M1_start': '', 'M2_start': ''}

        self.root_parent_list = [self.tree.nodes.get(self.tree.root)]
        self.parse_root_flag = False
        self.parse_mitosis_flag = {}
        self.parse_s_flag = {}
        self.smooth_flag = {}

        self.__division_count = 0

    def record_cell_division_count(self):
        self.__division_count += 1
        return self.__division_count

    def get_child(self, parent: CellNode) -> List[CellNode]:
        """Return all child nodes according to the parent node, including only direct child nodes"""
        return self.tree.children(parent.identifier)

    def search_root_node(self):
        """
        Find the root node of each generation of cells, that is, the first node of each branch of TrackingTree
        This method is to deal with the situation where mitosis occurs, that is, the tree begins to branch
        When the cell begins to undergo mitosis, the tree will split, record the node at this time, and backtrack
        as a generation of cells
        """
        branch_id = 0
        root_node = self.tree.nodes.get(self.tree.root)
        root_node.cell.set_cell_id(str(self.tree.track_id) + '_' + str(branch_id))
        self.lineage_dict[root_node] = {'cells': [root_node], 'branch_id': root_node.cell.branch_id, 'parent': None}
        if self.tree.root is None:
            return
        else:
            loop_queue = [root_node]
            while loop_queue:
                current = loop_queue.pop(0)
                ch = self.get_child(current)
                for next_node in ch:
                    loop_queue.append(next_node)
                if len(ch) > 1:
                    self.record_cell_division_count()
                    # if current not in self.lineage_dict or current in self.lineage_dict:  # not useful branch
                    for child_cell in ch:
                        self.lineage_dict[child_cell] = {'cells': [child_cell],
                                                         'branch_id': child_cell.cell.branch_id,
                                                         'branch_start': child_cell.cell.frame, 'parent': current}
                        if child_cell not in self.root_parent_list:
                            branch_id += 1
                            child_cell.cell.set_cell_id(str(self.tree.track_id) + '_' + str(branch_id))
                            self.root_parent_list.append(child_cell)
        self.parse_root_flag = True

    def parse_lineage(self, root_node):
        """Get the cell node sequence contained in each branch"""
        loop_queue = [root_node]
        last_node = None
        while loop_queue:
            current_node = loop_queue.pop(0)
            last_node = current_node
            ch = self.get_child(current_node)
            for loop_node in ch:
                loop_queue.append(loop_node)
            if current_node not in self.lineage_dict[root_node].get('cells'):
                self.lineage_dict[root_node].get('cells').append(current_node)
            if len(ch) > 1:
                break
        self.lineage_dict[root_node]['branch_end'] = last_node.cell.frame + 1

    def get_lineage_dict(self):
        if not self.parse_root_flag:
            self.search_root_node()
        for root in self.root_parent_list:
            self.parse_lineage(root)
            self.parse_mitosis_flag[root] = False
        return self.lineage_dict

    def bfs(self):
        """Traverse TrackingTree according to breadth first"""
        root_node = self.tree.nodes.get(self.tree.root)
        if self.tree.root is None:
            return
        else:
            loop_queue = [root_node]
            while loop_queue:
                current = loop_queue.pop(0)
                ch = self.get_child(current)
                for loop_node in ch:
                    loop_queue.append(loop_node)
                yield ch

    @staticmethod
    def check_mitosis_start(start_index, lineage_cell, area_size_t=0.9, mitosis_gap=20, m_predict_threshold=5):
        """
        Check the entry of the M period, if the area meets the conditions, check the next frame, if the area of the next
        frame is too small, it is not considered to have entered, and it is judged as a segmentation misjudgment.
        If the check is passed, check the next 6 frames. If the number of M phases is predicted to be greater than or
        equal to the threshold, it is judged to enter the M cell_type to pass, otherwise, the judgment fails.
        If the interval from the last entry to M is too short, it is also considered a misjudgment
        """
        predict_enter_cell = lineage_cell[start_index]
        next_cell = lineage_cell[start_index + 1]
        if lineage_cell[0].cell.cell_type == 'M' and start_index < mitosis_gap:
            return False
        elif next_cell.cell.area / predict_enter_cell.cell.area < area_size_t:
            return False
        else:
            predict_m_count = 0
            if len(lineage_cell) - start_index < 5:
                return True
            for i in range(min(8, len(lineage_cell) - start_index)):
                if lineage_cell[start_index + i].cell.cell_type == 'M':
                    predict_m_count += 1
            if predict_m_count >= m_predict_threshold:
                return True
            else:
                return False

    @staticmethod
    def check_ctype_start(type_name, start_index, linage_cell,
                          threshold=config.PROB_THRESHOLD * config.SMOOTH_WINDOW_LEN):
        """
        Check the entry of cell_type, if the cell is predicted to be in indicate cell_type, check 10 frames later,
        If the number of remaining frames is less than 10 frames, check all the remaining frames.
        If the cumulative predicted indicate cell type is greater than the threshold,
        If the judgment succeeds, otherwise, the judgment fails
        """
        count = 0
        if linage_cell[start_index].cell.cell_type == type_name:
            for i in range(min(config.SMOOTH_WINDOW_LEN, len(linage_cell) - start_index)):
                if linage_cell[start_index + i].cell.cell_type == type_name:
                    count += 1
        if count >= threshold:
            return True
        return False

    @staticmethod
    def check_ctype_exit(type_name, end_index, linage_cell, threshold=config.PROB_THRESHOLD * config.SMOOTH_WINDOW_LEN):
        """
        To judge the exit of cell_type, the judgment principle is the same as entering  cell_type,
        if the cells start to exit cell_type, check later
        """
        non_type_count = 0
        if linage_cell[end_index].cell.cell_type != type_name:
            for i in range(min(config.SMOOTH_WINDOW_LEN, len(linage_cell) - end_index)):
                if linage_cell[end_index + i].cell.cell_type != type_name:
                    non_type_count += 1
            if non_type_count >= threshold:
                return True
            return False

    def parse_mitosis(self, lineage: dict, root: CellNode, lineage_index=None):
        """
        Parse entry and exit of mitosis
        """
        area_size_t = 1.4
        cell_node_line = lineage.get('cells')
        mitosis_start_index = None
        exist_m_frame = 0
        for i in range(len(cell_node_line) - 1):
            before_cell_node = cell_node_line[i]
            current_cell_node = cell_node_line[i + 1]
            # print(f'{current_cell_node.cell.area / before_cell_node.cell.area:.2f}')
            if before_cell_node.cell.cell_type == 'M':
                exist_m_frame += 1
            if current_cell_node.cell.area / before_cell_node.cell.area >= area_size_t:
                if self.check_mitosis_start(i, cell_node_line,
                                            m_predict_threshold=config.PROB_THRESHOLD * config.SMOOTH_WINDOW_LEN):
                    mitosis_start_index = i + 1
                    break
            elif before_cell_node.cell.cell_type == 'M' and current_cell_node.cell.area / before_cell_node.cell.area < area_size_t:
                if self.check_mitosis_start(i, cell_node_line,
                                            m_predict_threshold=config.PROB_THRESHOLD * config.SMOOTH_WINDOW_LEN):
                    mitosis_start_index = i
                    break
        if mitosis_start_index is None:
            if len(cell_node_line) < 5:
                if exist_m_frame >= 3:
                    for cell_node in cell_node_line:
                        cell_node.cell.cell_type = 'M'
                    lineage['m2_start'] = 0
            else:
                if lineage_index != 0:
                    for cell_node in cell_node_line[: 3]:
                        cell_node.cell.cell_type = 'M'
                    lineage['m1_start'] = 0
        else:
            for m_index in range(mitosis_start_index, len(cell_node_line)):
                cell_node_line[m_index].cell.cell_type = 'M'
            lineage['m2_start'] = mitosis_start_index
        self.parse_mitosis_flag[root] = True

    def parse_s(self, lineage: dict, root: CellNode, lineage_index=None):
        """To judge the entry of S cell_type"""
        cell_node_line = lineage.get('cells')
        s_start_index = None
        s_exit_index = None
        if not self.parse_mitosis_flag[root]:
            self.parse_mitosis(lineage, root, lineage_index=lineage_index)
        if lineage.get('m1_start') is not None:
            check_start = lineage.get('m1_start')
        else:
            check_start = 0
        for cell_node_index in range(check_start, len(cell_node_line)):
            if cell_node_line[cell_node_index].cell.cell_type == 'S':
                if self.check_ctype_start('S', cell_node_index, cell_node_line):
                    lineage['s_start'] = s_start_index = cell_node_index
                    break
        if s_start_index is not None:
            for cell_node_index_2 in range(s_start_index, len(cell_node_line)):
                if cell_node_line[cell_node_index_2].cell.cell_type != 'S':
                    if self.check_ctype_exit('S', cell_node_index_2, cell_node_line):
                        lineage['s_exit'] = s_exit_index = cell_node_index_2
                        break

        if s_start_index is not None:
            if s_exit_index is not None:
                end = s_exit_index
            else:
                end = len(cell_node_line)
            for cell_node_index_s in range(s_start_index, end):
                cell_node_line[cell_node_index_s].cell.cell_type = 'S'
        self.parse_s_flag[root] = True

    def parse_g1_g2(self, lineage: dict, root: CellNode, lineage_index=None):
        """Accurately distinguish G1/G2 into G1, G2"""
        cell_node_line = lineage.get('cells')
        g1_start_index = None
        g1_exit_index = None
        g2_start_index = None
        g2_exit_index = None
        m1_start = lineage.get('m1_start')
        m2_start = lineage.get('m2_start')
        if not self.parse_s_flag[root]:
            self.parse_s(lineage, root)
        if lineage.get('s_start') is not None:
            # 1. track starts from  S
            # 2. track starts from G1 cell_type
            # 3. The track starts from the M1
            g1_exit_index = lineage.get('s_start')
            if m1_start is not None:
                g1_start_index = 3
            else:
                g1_start_index = 0
            if lineage.get('s_exit') is not None:
                g2_start_index = lineage.get('s_exit')
                if m2_start is not None:
                    g2_exit_index = m2_start
                else:
                    g2_exit_index = len(cell_node_line)

        else:
            # cells are not in S cell_type
            # 1. The track starts from the M1 period and does not enter the S cell_type
            # 2. The track starts from the G1 cell_type and does not enter the S cell_type
            # 3. The track starts from the G2 period and enters the M2 cell_type
            # 3. The track starts from the G2 period and does not enter the M2 cell_type
            if m2_start is not None:
                if len(cell_node_line) > 5:
                    g2_start_index = 0
                    g2_exit_index = m2_start
            elif m1_start is not None:
                g1_start_index = 3
                g1_exit_index = len(cell_node_line)
            else:
                if len(cell_node_line) > 10:
                    if lineage_index == 0:
                        g1_start_index = 0
                        g1_exit_index = len(cell_node_line)
                    else:
                        g2_start_index = 0
                        g2_exit_index = len(cell_node_line)
                else:
                    g1_start_index = 0
                    g1_exit_index = len(cell_node_line)

        if g1_start_index is not None:
            lineage['g1_start'] = g1_start_index
            if g1_exit_index is not None:
                end = g1_exit_index
            else:
                end = len(cell_node_line)
            for cell_node_index_g1 in range(g1_start_index, end):
                cell_node_line[cell_node_index_g1].cell.cell_type = 'G1'
        if g2_start_index is not None:
            lineage['g2_start'] = g2_start_index
            if g2_exit_index is not None:
                end_2 = g2_exit_index
            else:
                end_2 = len(cell_node_line)
            for cell_node_index_g2 in range(g2_start_index, end_2):
                cell_node_line[cell_node_index_g2].cell.cell_type = 'G2'

    def parse_mitosis_error(self, lineage: dict, root: CellNode, lineage_index=None):
        """
        If the two daughter cells after division are not matched, all subsequent cells will be in the M cell_type.
        At this time, it should be corrected to G1 within a certain period of time.
        """
        cell_node_line = lineage.get('cells')
        m1_start = lineage.get('m1_start')
        m2_start = lineage.get('m2_start')
        m_count = 30
        if m1_start is not None:
            m_start = m1_start
        elif m2_start is not None:
            m_start = m2_start
        else:
            return
        for index in range(m_start, len(cell_node_line)):
            if cell_node_line[index].cell.cell_type == 'M':
                m_count -= 1
                if m_count < 0:
                    cell_node_line[index].cell.cell_type = 'G1'

    def set_cell_id(self, lineage: dict, root: CellNode, lineage_index):
        cell_node_line = lineage.get('cells')
        branch_id = lineage_index
        cell_id = str(self.tree.track_id) + '_' + str(branch_id)
        if lineage_index == 0:
            root.cell.set_cell_id(cell_id)
            root.cell.set_branch_id(branch_id)
            lineage['parent'] = root
        else:
            parent = lineage['parent']
            parent.cell.set_cell_id(parent.cell.cell_id)
            # parent.cell.set_cell_id(cell_id)
            # parent.cell.set_branch_id(branch_id)
        for cell_node in cell_node_line:
            cell_node.cell.set_cell_id(cell_id)
            cell_node.cell.set_branch_id(branch_id)
            cell_node.cell.set_track_id(self.tree.track_id, 1)

    def parse_lineage_phase(self, lineage: dict, root: CellNode, linage_index):
        root = root
        self.parse_mitosis(lineage, root, linage_index)
        self.parse_s(lineage, root, linage_index)
        self.parse_g1_g2(lineage, root, linage_index)
        self.parse_mitosis_error(lineage, root, linage_index)
        self.set_cell_id(lineage, root, linage_index)

    def parse_lineage_branch_id(self, lineage, branch_id):
        pass

    def smooth_type(self, cell_lineage, root, lineage_index):
        if N_CLASS and CLASS_NAME:
            if len(CLASS_NAME) != N_CLASS:
                logging.error('The number of category names and category numbers is not equal, '
                              'try to change N_CLASS OR CLASS_NAME')
                return
        if CLASS_NAME:
            if CLASS_NAME == ['G1', 'S', 'G2', 'M']:
                self.parse_lineage_phase(cell_lineage, root, lineage_index)
                return
            else:
                class_name = CLASS_NAME
        elif N_CLASS:
            class_name = set()
            for cell in cell_lineage.get('cells'):
                class_name.add(cell.cell.cell_type)
        else:
            logging.info('N_CLASS and CLASS_NAME not provided, ignore the whole process')
            return
        resolved_index_map = {}
        for ctype in class_name:
            cell_node_line = cell_lineage.get('cells')
            check_start = 0
            start_index = None
            exit_index = None
            for cell_node_index in range(check_start, len(cell_node_line)):
                if cell_node_line[cell_node_index].cell.cell_type == ctype:
                    if self.check_ctype_start(ctype, cell_node_index, cell_node_line):
                        cell_lineage[f'{ctype}_start'] = start_index = cell_node_index
                        break
            if start_index is not None:
                for cell_node_index_2 in range(start_index, len(cell_node_line)):
                    if cell_node_line[cell_node_index_2].cell.cell_type != ctype:
                        if self.check_ctype_exit(ctype, cell_node_index_2, cell_node_line):
                            cell_lineage[f'{ctype}_exit'] = exit_index = cell_node_index_2 + 1
                            break
            if start_index is not None:
                if exit_index is not None:
                    end = exit_index
                else:
                    end = len(cell_node_line)
                resolved_index_map[ctype] = start_index
                for cell_node_index in range(start_index, end):
                    cell_node_line[cell_node_index].cell.cell_type = ctype
        if 0 not in resolved_index_map.values():
            if not resolved_index_map:
                type_count = []
                for cell in cell_lineage.get('cells'):
                    type_count.append(cell.cell.cell_type)
                counter = Counter(type_count)
                resolved_type = counter.most_common(1)[0][0]
                for cell in cell_lineage.get('cells'):
                    cell.cell.cell_type = resolved_type
            else:
                resolved_type = min(resolved_index_map, key=resolved_index_map.get)
                for cell in cell_lineage.get('cells')[:min(resolved_index_map.values())]:
                    cell.cell.cell_type = resolved_type
        self.smooth_flag[root] = True


def pares_single_tree(tree: TrackingTree):
    parser = TreeParser(tree)
    parser.search_root_node()
    parser.get_lineage_dict()
    for node_index in range(len(parser.root_parent_list)):
        cell_lineage = parser.lineage_dict.get(parser.root_parent_list[node_index])
        if tree.get_node(tree.root).cell.cell_type is None:
            parser.set_cell_id(cell_lineage, root=parser.root_parent_list[node_index], lineage_index=node_index)
        else:
            # parser.parse_lineage_phase(cell_lineage, root=parser.root_parent_list[node_index], linage_index=node_index)
            parser.smooth_type(cell_lineage, root=parser.root_parent_list[node_index], lineage_index=node_index)
            parser.set_cell_id(cell_lineage, root=parser.root_parent_list[node_index], lineage_index=node_index)
            # print(cell_lineage)
    return parser


def paste(tree_pre: TrackingTree, paste_node: CellNode, tree_next: TrackingTree):
    """relink two tree"""
    tree_pre.paste(paste_node.nid, tree_next)

def find_combinations_objects(A, B, gap_threshold):
    # A is the root node list
    # B is the leaves node list
    result = []
    b_frames_set = {obj.cell.frame for obj in B}
    for a_obj in A:
        for diff in range(1, gap_threshold + 1):
            if a_obj.cell.frame - diff in b_frames_set:
                try:
                    result.append([a_obj, next(obj for obj in B if obj.cell.frame == a_obj.cell.frame + diff)])
                except StopIteration:
                    pass
    return result

def relink_tree(tracker, gap_threshold=3):
    """When the frame index of the root node of a tree is greater than the frame index of the leaf node of another tree,
    and the difference between the two is within a certain range(gap_threshold), it will judged whether the two trees can
    be relinked,according to the similarity of this two cell"""
    root_nodes = {}
    leaves_nodes = {}
    for tree in tracker.trees:
        root_nodes[tree.get_node(tree.root)] = tree
        for leaf in tree.leaves():
            leaves_nodes[leaf] = tree
    link_candidates = find_combinations_objects(root_nodes.keys(), leaves_nodes.keys(), gap_threshold)
    # Note: this function maybe import new bug, or make the performance bad.
    if link_candidates:
        for pair in link_candidates:
            if tracker.matcher.match_similar(pair[0].cell, pair[1].cell).get('IoU') > 0:
                paste(leaves_nodes[pair[1]], pair[1], root_nodes[pair[0]])


def run_track(annotation, track_range=None, dic=None, mcy=None, speed_filename=None):
    tracker = Tracker(annotation, dic=dic, mcy=mcy)
    if track_range:
        tracker.track(range=track_range, speed_filename=speed_filename)
    else:
        tracker.track()
    parser_dict = {}
    for tree in tracker.trees:
        parser = pares_single_tree(tree)
        parser_dict[tree] = parser
    tracker.parser_dict = parser_dict
    relink_tree(tracker)
    return tracker


def track_tree_to_table(tracker: Tracker, filepath):
    """Export track result to table"""
    track_detail_columns = ['frame_index', 'track_id', 'cell_id', 'parent_id', 'center_x', 'center_y', 'cell_type',
                            'mask_of_x_points', 'mask_of_y_points']
    track_detail_dataframe = pd.DataFrame(columns=track_detail_columns)

    def generate_series(cell_lineage):
        cell_nodes = cell_lineage.get('cells')
        parent = cell_lineage.get('parent')
        series_list = []
        new_nodes = []
        before_index = 0
        current_index = 1
        if len(cell_nodes) == 1:
            new_nodes.append(cell_nodes[0])
        for cell_node_index in range(len(cell_nodes) - 1):
            before_node = cell_nodes[before_index]
            current_node = cell_nodes[current_index]
            new_nodes.append(before_node)
            if current_node.cell.frame - before_node.cell.frame != 1:
                for frame in range(before_node.cell.frame + 1, current_node.cell.frame):
                    # gap_cell = deepcopy(before_node.cell)
                    # gap_cell.frame = frame
                    gap_cell = Cell(position=before_node.cell.position, cell_type=before_node.cell.cell_type,
                                    frame_index=frame)
                    gap_cell.set_track_id(before_node.cell.track_id, 1)
                    gap_cell.set_cell_id(before_node.cell.cell_id)
                    gap_cell.set_branch_id(before_node.cell.branch_id)
                    gap_node = CellNode(gap_cell)
                    new_nodes.append(gap_node)
            elif current_index == len(cell_nodes) - 1:
                new_nodes.append(current_node)
            else:
                pass
            before_index += 1
            current_index += 1

        # for node in cell_nodes:
        # print(new_nodes)
        for node in new_nodes:
            col = [node.cell.frame, node.cell.track_id,
                   node.cell.cell_id, parent.cell.cell_id,
                   node.cell.center[0], node.cell.center[1],
                   node.cell.cell_type,
                   node.cell.position[0], node.cell.position[1]]
            s = pd.Series(dict(zip(track_detail_columns, col)))
            series_list.append(s)
        return series_list, new_nodes

    parser_dict = tracker.parser_dict
    for tree in parser_dict:
        if tree.track_id == 4:
            p = parser_dict[tree]
        parser = parser_dict[tree]
        for node_index in parser.root_parent_list:
            cell_lineage = parser.lineage_dict.get(node_index)
            series_list, new_node_list = generate_series(cell_lineage)
            for series in series_list:
                track_detail_dataframe = track_detail_dataframe._append(series, ignore_index=True)
    fname = filepath
    track_detail_dataframe.to_csv(fname, index=False)


def track_trees_to_json(tracker: Tracker, output_fname, xrange, basename=None):
    """Export track result to json file"""
    if basename is None:
        prefix = 'mcy'
    else:
        prefix = basename

    def update_region(node):
        # print(type(node))
        region = node.cell.region
        cell_type = node.cell.cell_type
        track_id = node.cell.track_id
        cell_id = node.cell.cell_id
        region['region_attributes']['phase'] = cell_type
        region['region_attributes']['track_id'] = track_id
        region['region_attributes']['cell_id'] = cell_id
        region['region_attributes']['id'] = track_id
        node.cell.set_region(region)

    result = {}
    frame_map = {}

    for frame in range(xrange):
        image_name = prefix + '-' + str(frame).zfill(4) + '.png'
        tmp = {
            image_name: {
                "filename": image_name,
                "size": int(RAW_INPUT_IMAGE_SIZE[0] * RAW_INPUT_IMAGE_SIZE[1]),
                "regions": [],
                "file_attributes": {}
            }
        }
        result[image_name] = deepcopy(tmp[image_name])
        frame_map[frame] = image_name
        del tmp
    for tree in tracker.trees:

        for node in tree.all_nodes():
            frame = node.cell.frame
            update_region(node)
            region = node.cell.region
            image_name = frame_map.get(frame)
            if image_name is not None:
                tmp_frame = result.get(image_name)
                tmp_frame['regions'].append(region)
    with open(output_fname, 'w') as f:
        json.dump(result, f)


def track_tree_to_TRA(tracker: Tracker, output_fname=None, xrange=None, basename=None):
    """Export track result as TRA format"""

    TRA = []
    parser_dict = tracker.parser_dict
    for tree in parser_dict:
        parser = parser_dict[tree]
        for node_index in parser.root_parent_list:
            cell_lineage = parser.lineage_dict.get(node_index)
            cells = cell_lineage.get('cells')
            parent = cell_lineage['parent']
            branch_id = cell_lineage['branch_id']
            L = int(cells[0].cell.cell_id.replace('_', ''))
            B = cells[0].cell.frame
            E = cells[-1].cell.frame
            if branch_id == 0:
                P = 0
            else:
                P = int(parent.cell.cell_id.replace('_', ''))
            line = f'{L} {B} {E} {P}\n'
            TRA.append(line)
    if output_fname:
        with open(output_fname, 'w') as f:
            f.writelines(TRA)


def track_tree_to_mask(tracker, width, height, output_dir):
    parser_dict = tracker.parser_dict
    cell_each_frame = {}  # {frame: cell_list}
    for tree in parser_dict:
        parser = parser_dict[tree]
        for node_index in parser.root_parent_list:
            cell_lineage = parser.lineage_dict.get(node_index)
            cells = cell_lineage.get('cells')
            for cell in cells:
                if cell.cell.frame not in cell_each_frame:
                    cell_each_frame[cell.cell.frame] = {cell}
                else:
                    cell_each_frame[cell.cell.frame].add(cell)
    if not os.path.exists(output_dir):
        os.mkdir(output_dir)
    for frame in cell_each_frame.keys():
        fname = os.path.join(output_dir, f"mask" + f"{frame}".zfill(3) + ".tif")
        mask_arr = np.zeros((height, width), dtype=np.uint16)
        for cell in cell_each_frame[frame]:
            contours = cell.cell.contours
            L = int(cell.cell.cell_id.replace('_', ''))
            cv2.fillConvexPoly(mask_arr, contours, (L))
        tifffile.imwrite(fname, mask_arr)


def run(annotation, output_dir, basename, track_range=None, save_visualize=True, visualize_background_image=None,
        dic=None, mcy=None, track_to_json=True):
    if track_range is None:
        if type(annotation) is str:
            with open(annotation, encoding='utf-8') as f:
                data = json.load(f)
        elif type(annotation) is dict:
            data = annotation
        else:
            raise ValueError(f"annotation type error {type(annotation)}")
        xrange = len(data)
    else:
        xrange = track_range + 2
    if config.RECORD_SPEED:
        speed_output_filename = os.path.join(output_dir, 'track_speed.csv')
    else:
        speed_output_filename = None
    tracker = run_track(annotation, track_range=xrange - 2, dic=dic, mcy=mcy, speed_filename=speed_output_filename)
    track_table_fname = os.path.join(output_dir, 'track.csv')
    track_visualization_fname = os.path.join(output_dir, 'track_visualization.tif')
    track_json_fname = os.path.join(output_dir, 'result_with_track.json')
    tracktree_save_path = os.path.join(output_dir, 'TrackTree')
    track_tree_to_table(tracker, track_table_fname)
    tracker.track_tree_to_json(tracktree_save_path)
    # track_tree_to_TRA(tracker, os.path.join(output_dir, 'TRA.txt'))
    # image_width, image_height = imagesize.get(mcy)
    # track_tree_to_mask(tracker, image_width, image_height, os.path.join(output_dir, 'mask'))
    if track_to_json:
        track_trees_to_json(tracker, track_json_fname, xrange=xrange, basename=basename)
    if save_visualize:
        tracker.visualize_to_tif(visualize_background_image, track_visualization_fname, tracker.trees, xrange=xrange)


if __name__ == '__main__':
    # annotation = r'G:\杂项\example\example-annotation.json'
    # mcy_img = r'G:\杂项\example\example-image.tif'
    # dic_img = r'G:\杂项\example\example-bf.tif'
    annotation = r"G:\raw data\demo\annotation.json"
    mcy_img = r"G:\raw data\demo\image.tif"
    dic_img = r"G:\CTC dataset\Fluo-N2DL-HeLa\Fluo-N2DL-HeLa\01_ST\SEG.tif"
    # tracker = run_track(r'G:\paper\evaluate_data\copy_of_1_xy10\result-GT.json', track_range=10)
    # tracker = run_track(annotation, track_range=30)
    # background_filename_list = [os.path.join(r'G:\paper\evaluate_data\copy_of_1_xy10\tif-seq', i) for i in
    #                             os.listdir(r'G:\paper\evaluate_data\copy_of_1_xy10\tif-seq')]
    # print(background_filename_list)
    # for i in tracker.trees:
    #     print(i)
    # tracker.visualize_single_tree(tree=tracker.trees[51],
    #                               save_dir=r'G:\paper\evaluate_data\copy_of_1_xy10\single_tree_visualize',
    #                               background_filename_list=background_filename_list, xrange=10)

    run(annotation=annotation, output_dir=r"G:\raw data\demo\test",
        track_range=None, dic=None, mcy=mcy_img,
        save_visualize=True, visualize_background_image=mcy_img,
        track_to_json=True, basename=r'man_seg')



from __future__ import annotations

import json

import numpy as np
import cv2
from typing import Tuple, List
import matplotlib.pyplot as plt
import openpyxl as px
import tifffile
from SCTrack.prepare import convert_dtype


class Stack(object):
    def __init__(self):
        self.start_frame = None
        self.end_frame = None
        self.center_info = None
        self.id = None

    def __len__(self):
        if self.start_frame is None:
            return 0
        else:
            # return self.end_frame - self.start_frame + 1
            return len(self.center_info)

    def __str__(self):
        return f"start frame: {self.start_frame}, end frame: {self.end_frame}, total: " \
               f"{self.end_frame - self.start_frame + 1}, len: {len(self.center_info)} " \
               f"\ncenter info: {self.center_info}"

    def __repr__(self):
        return self.__str__()


class RefinedParser(object):
    def __init__(self, path):
        wb = px.load_workbook(path)
        sheet = wb[wb.sheetnames[0]]
        self.frame_details = sheet['A'][1:]
        self.id_details = sheet['B'][1:]
        self.lineage_details = sheet['C'][1:]
        self.parent_id_details = sheet['D'][1:]
        self.phase_details = sheet['S'][1:]
        self.center0 = sheet['F'][1:]
        self.center1 = sheet['E'][1:]

    def parse_id(self):
        id_record = []
        id_info = []
        current_index = 0
        for i in range(len(self.id_details)):
            if self.id_details[i].value not in id_record:
                id_record.append(self.id_details[i].value)
        for _id in id_record:
            length = 0
            start = current_index
            for j in self.id_details[current_index:]:
                if j.value == _id:
                    length += 1
                else:
                    current_index += length
                    break
            end = start + length - 1
            id_info.append({'id': _id, 'start': start, 'end': end, 'continue': length})
            # print({'id': _id, 'start': start, 'end': end, 'continue': length})
        return id_info

    def parse_position(self):
        id_info = self.parse_id()
        position = []
        for i in id_info:
            _id = i['id']
            start_index = i['start']
            end_index = i['end']
            # print(_id, start_index, end_index, self.frame_details[start_index].value,
            #       self.frame_details[end_index].value)
            position.append({'id': _id, 'start_index': start_index, 'end_index': end_index,
                             'start_frame': self.frame_details[start_index].value,
                             'end_frame': self.frame_details[end_index].value})
        return position

    def get_stack(self):
        positions = self.parse_position()
        stacks = []
        for i in positions:
            stack = Stack()
            stack.start_frame = i.get('start_frame')
            stack.end_frame = i.get('end_frame')
            stack.center_info = []
            stack.id = i.get('id')
            for index in range(i.get('start_index'), i.get('end_index') + 1):
                stack.center_info.append((round(self.center0[index].value), round(self.center1[index].value)))
            stacks.append(stack)
        return stacks


class JsonParser(object):
    def __init__(self, file):
        with open(file) as f:
            self.data = json.load(f)
        self.index_map = {}
        __index = 0
        for frame in self.data:
            self.index_map[__index] = frame
            __index += 1

    def __len__(self):
        return len(self.index_map)

    def parse_json(self):
        all_coords = {}
        for frame in self.data:
            regions = self.data[frame]['regions']
            coords = []
            for j in regions:
                all_x = j['shape_attributes']['all_points_x']
                all_y = j['shape_attributes']['all_points_y']
                coords.append((all_x, all_y))
            all_coords[frame] = coords
        return all_coords

    def get_coords_by_frame_name(self, frame_name):
        regions = self.data[frame_name]['regions']
        coords = []
        for j in regions:
            all_x = j['shape_attributes']['all_points_x']
            all_y = j['shape_attributes']['all_points_y']
            coords.append((all_x, all_y))
        return coords

    def get_coords_by_frame_index(self, frame_index=None):
        frame_name = self.index_map[frame_index]
        return self.get_coords_by_frame_name(frame_name)

    def get_frame_name_by_index(self, index):
        return self.index_map[index]


class Mask(object):
    def __init__(self, mask=None, center=None, coord=None):
        self._mask = mask
        self._center = center
        self.id = None
        self.frame_index = None
        self.coord = coord

    @property
    def mask(self):
        if self._mask is None:
            raise ValueError("No mask")
        return self._mask

    @property
    def center(self) -> Tuple[int | float]:
        if self._center is None:
            raise ValueError("No available center position")
        return self._center

    def __str__(self):
        return f"Mask object at center of {self._center[0], self._center[1]}"

    def __repr__(self):
        return self.__str__()


def coordinate2mask(coords: np.ndarray | list | tuple, value: int = 255, image_size: Tuple[int, int] = None) -> \
        List[Mask]:
    results = []
    for coord in coords:
        if image_size is None:
            mask = np.zeros((2048, 2048), dtype=np.uint8)
        else:
            mask = np.zeros(image_size, dtype=np.uint8)
        points = []
        for j in range(len(coord[0])):
            x = int(coord[0][j])
            y = int(coord[1][j])
            points.append((x, y))
        contours = np.array(points)
        cv2.fillConvexPoly(mask, contours, (value, 0, 0))
        mMask = Mask(mask, center=(round(float(np.mean(coord[0]))), round(float(np.mean(coord[1])))), coord=coord)
        results.append(mMask)
    return results


def coord2counter(coord):
    points = []
    for j in range(len(coord[0])):
        x = int(coord[0][j])
        y = int(coord[1][j])
        points.append((x, y))
    contours = np.array(points)
    return contours


def link(json_file, refined_file):
    ref_file = refined_file
    refined = RefinedParser(ref_file)
    stacks = refined.get_stack()
    json_parser = JsonParser(json_file)
    for stack in stacks:
        linked_masks = []
        centers = stack.center_info
        start_frame = stack.start_frame
        for index in range(len(stack.center_info)):
            center = centers[index]
            coords = json_parser.get_coords_by_frame_index(index + start_frame)
            masks = coordinate2mask(coords)
            for m in masks:
                if (abs(center[0] - m.center[0]) <= 10) and (abs(center[1] - m.center[1]) <= 10):
                    # print(f"id index: {index} detected! at frame: {index + start_frame}.")
                    m.frame_index = index + start_frame
                    linked_masks.append(m)
                    break
        yield linked_masks


def extractRoiFromImg(images: str | np.ndarray, mask: np.ndarray) -> np.ndarray:
    """
    Extract the area in the original image according to the mask. Note that it can only be a single-channel image.
    If it is rgb, please convert it to a grayscale image first.
    :param images: original images
    :param mask: mask file
    :return: single cell image data
    """
    if type(images) is str:
        src = cv2.imread(images, -1)
    else:
        src = images
    new_src = convert_dtype(src)
    dst = np.zeros_like(new_src, dtype=np.uint8)
    cv2.copyTo(new_src, mask, dst)
    return dst


def csv2mask(jsonfile, excelfile, mask_filename):
    jp = JsonParser(jsonfile)
    rp = RefinedParser(excelfile)
    stacks = rp.get_stack()
    frame_info = {}
    for index in range(len(jp)):
        frame_info[index] = set()
    print(frame_info)
    for stack in stacks:
        start = stack.start_frame
        for i in range(len(stack)):
            frame_info[start + i].add((stack.center_info[i], stack.id))
    all_masks = []
    for i in range(len(frame_info)):
        coords = jp.get_coords_by_frame_index(i)
        masks = np.zeros((2048, 2048), dtype=np.uint8)
        match_count = 0
        matched_center = []
        for j in frame_info[i]:
            for coord in coords:
                if (abs(j[0][0] - round(float(np.mean(coord[0])))) < 8 and
                        abs(j[0][1] - round(float(np.mean(coord[1])))) < 8):
                    ret = coordinate2mask([coord], j[1])
                    matched_center.append((round(float(np.mean(coord[0]))), round(float(np.mean(coord[1])))))
                    match_count += 1
                    masks += ret[0].mask
        for coord in coords:
            if (round(float(np.mean(coord[0]))), round(float(np.mean(coord[1])))) not in matched_center:
                e_ret = coordinate2mask([coord], 255)
                masks += e_ret[0].mask
        plt.imshow(masks, cmap='gray')
        plt.show()
        print(match_count)
        all_masks.append(masks)
    all_masks = np.array(all_masks)
    tifffile.imsave(mask_filename.replace('.tif', '-1.tif'), all_masks[:all_masks.shape[0] // 2])
    tifffile.imsave(mask_filename.replace('.tif', '-2.tif'), all_masks[all_masks.shape[0] // 2:])


if __name__ == '__main__':
    pass



#! /usr/bin/python3
# -*- coding: utf-8 -*-
# @FileName: setup.py
# @Author: Li Chengxin 
# @Time: 2023/7/4 13:43

import setuptools

with open('readme.md', 'r', encoding='utf-8') as f:
    long_description = f.read()

with open('requirements.txt', 'r') as fr:
    pkg_requirements = fr.read().split('\n')
    pkg_requirements.remove('')


VERSION = '0.0.9'

setuptools.setup(
    name='SC-Track',
    author="Li Chengxin",
    author_email="914814442@qq.com",
    url="https://github.com/frozenleaves/SC-Track",
    license="GNU General Public License v3.0",
    version=VERSION,
    description='single cell tracking package',
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=setuptools.find_packages(),
    zip_safe=False,
    python_requires=">=3.7",
    install_requires=pkg_requirements,
    # package_data={'SCTrack': ['notebook/*']},
    # include_package_data=True,
    entry_points={
        'console_scripts': [
            'sctrack = SCTrack.sctrack:main',
        ],
    },
)



#! /usr/bin/python3
# -*- coding: utf-8 -*-
# @FileName: utils.py
# @Author: Li Chengxin 
# @Time: 2023/6/30 22:15

from __future__ import annotations

import logging
import sys
import time
import os
from copy import deepcopy
from functools import wraps
import skimage.exposure as exposure
from skimage.util import img_as_ubyte
from libtiff import TIFF
import tifffile
import matplotlib.pyplot as plt
import numpy as np
import cv2
import glob
from tqdm import tqdm
from SCTrack import template

PHASE_MAP = {
    "G1/G2": 0,
    "M": 1,
    "S": 2
}


def time_it(logger: logging.Logger = None, using=False):
    """
    Decorator used to record the execution time of a function, using a log to output the execution time
    :param logger: Python logger
    :param using: bool flag, whether the decorator function enabled.
    :return: decorator
    """

    def switch(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            if logger is None or using is False:
                func(*args, **kwargs)
            else:
                start = time.time()
                func(*args, **kwargs)
                end = time.time()
                logger.info(f'{func} cost time:  {end - start:.4f}s')

        return wrapper

    return switch


def tif2png(img: str | os.PathLike, png_dir, gamma=0.1):
    """
    :param img: TIF image filepath
    :param png_dir: storing pngs dir
    :param gamma: image γ coefficient
    :return: None
    """
    tif = TIFF.open(img, )
    index = 0
    if not os.path.exists(png_dir):
        os.makedirs(png_dir)
    for i in tif.iter_images():
        filename = os.path.basename(img).replace('.tif', '-' + str(index).zfill(4) + '.png')
        img_mcy = exposure.adjust_gamma(i, gamma)
        png = img_as_ubyte(img_mcy)
        plt.imsave(os.path.join(png_dir, filename), png, cmap='gray')
        index += 1


def readTif(filepath):
    """
    A generator, from a multi-frame tif file, read frame by frame and return the image and filename of each frame.

    """
    tif = tifffile.TiffFile(filepath)
    num = len(tif.pages)
    if num <= 1:
        tif.close()
        tif = tifffile.imread(filepath)
        for i, frame in enumerate(tif):
            filename = os.path.basename(filepath).replace('.tif', '-' + str(i).zfill(4) + '.tif')
            yield frame, filename
    else:
        for i in range(num):
            frame = tif.pages[i].asarray()
            filename = os.path.basename(filepath).replace('.tif', '-' + str(i).zfill(4) + '.tif')
            yield frame, filename


def convert_dtype(__image: np.ndarray) -> np.ndarray:
    """
    Convert image format from uint16 to uint8
    :param __image: uint16 image ndarray
    :return: uint8 image ndarray
    """
    min_16bit = np.min(__image)
    max_16bit = np.max(__image)
    image_8bit = np.array(np.rint(255 * ((__image - min_16bit) / (max_16bit - min_16bit))), dtype=np.uint8)
    return image_8bit


def mask_to_coords(mask: np.ndarray | os.PathLike, filename):
    """
    Convert mask to contour coordinates. For the mask of an image, each different instance requires the same and unique
    pixel values within its range, and the pixel values of different instances are different. This requirement is only
    for instances within the same frame of image.

    :param mask: mask grayscale image, np.ndarray
    :param filename: filename for the mask
    :return: a region tmp dict for each mask, the format see template.py
    """
    if type(mask) is str:
        mask = tifffile.imread(mask)
    mask_values = np.delete(np.unique(mask), np.where(np.unique(mask) == 0))
    frame_tmp = deepcopy(template.FRAME_TMP)
    regions = []
    for i in mask_values:
        new = np.where(mask == i, 1, 0)
        new = new.astype(np.uint8)
        contours, hierarchy = cv2.findContours(new, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_NONE)
        x_coords = []
        y_coords = []
        for contour in contours:
            for pt in contour:
                x_coords.append(int(pt[0][0]))
                y_coords.append(int(pt[0][1]))
        region = deepcopy(template.REGIONS_TMP)
        region["shape_attributes"]["all_points_x"] = x_coords
        region["shape_attributes"]["all_points_y"] = y_coords
        regions.append(region)
    frame_tmp["regions"] = regions
    if os.path.splitext(filename)[1] == '.tif':
        filename = filename.replace('.tif', '.png')
    frame_tmp["filename"] = filename
    assert len(mask.shape) == 2
    frame_tmp["size"] = mask.shape[0] * mask.shape[1]
    return frame_tmp


def mask_seq_to_json(mask_dir, xrange=None):
    """
    Convert mask sequence files to json annotation files.
    :param mask_dir: The folder path for the mask image sequence location
    :param xrange: The number of conversions, counting from the beginning of the mask sequence.
    :return: json annotation dict, can be directly dump to the json file.
    """
    tif_files = glob.glob(os.path.join(mask_dir, '*.tif'))
    json_result = {}
    if xrange is None:
        xrange = len(tif_files)
    elif xrange > len(tif_files) - 1:
        xrange = len(tif_files)
    for file in tqdm(tif_files[:xrange + 1], desc='convert process'):
        filename = os.path.basename(file).replace('.tif', '.png')
        ret = mask_to_coords(file, filename)
        json_result[filename] = ret
    return json_result


def mask_tif_to_json(image, xrange=None):
    """
    Convert Multi-frame TIF mask image to json annotation files.
    :param image: The TIF filepath for mask location
    :param xrange: The number of conversions, counting from the beginning of the mask.
    :return: json annotation dict, can be directly dump to the json file.
    """
    if not (image.endswith('.tif') or image.endswith('.tiff')):
        logging.error(f'image {image} must be a tif/tiff file !')
        sys.exit(-1)
    json_result = {}
    count = 0
    with tifffile.TiffFile(image) as tif:
        num_frames = len(tif.pages)
    if xrange is None or xrange > num_frames:
        it = num_frames
    else:
        it = xrange
    for img, frame_name in tqdm(readTif(image), total=it, desc='convert process'):
        filename = frame_name.replace('.tif', '.png')
        ret = mask_to_coords(img, filename)
        json_result[filename] = ret
        if xrange:
            if count >= xrange:
                break
        count += 1
    return json_result


def mask_to_json(annotation: 'file or folder', xrange=None):
    """
     Convert  mask image to json annotation files

    :param annotation: mask filepath.
    :param xrange: The number of conversions, counting from the beginning of the annotation sequence.
    :return: json annotation dict, can be directly dump to the json file.
    """
    if os.path.isdir(annotation):
        return mask_seq_to_json(annotation, xrange)
    elif os.path.isfile(annotation):
        return mask_tif_to_json(annotation, xrange)


if __name__ == '__main__':
    # filedir = r'G:\CTC dataset\Fluo-N2DL-HeLa\Fluo-N2DL-HeLa\01_ST\SEG'
    # jsons = mask_seq_to_json(filedir)
    # with open(r'G:\CTC dataset\Fluo-N2DL-HeLa\Fluo-N2DL-HeLa\01_ST\test.json', 'w') as f:
    #     json.dump(jsons, f)
    mask_tif_to_json(r'G:\paper\evaluate_data\copy_of_1_xy10\mcy.tif')



#! /usr/bin/python3
# -*- coding: utf-8 -*-
# @FileName: config.py.py
# @Author: Li Chengxin 
# @Time: 2023/6/30 22:12


# Global configuration variables. If you are not sure what these parameters are, please do not modify them,
# except for those marked with *

from __future__ import annotations

from typing import List, Tuple

AVAILABLE_RANGE_COEFFICIENT: float = 1.5 # The available range coefficient of candidate screening for matching algorithms

RAW_INPUT_IMAGE_SIZE: Tuple = (2048, 2048)   # * image size (width, height)

GAP_WINDOW_LEN = 100 # Maximum number of lost track

ENTER_DIVISION_THRESHOLD = 50 # Starting from the completion of cell division, the second division is not allowed until the value is exceeded

USING_IMAGE_FOR_TRACKING = False   # * Using additional image information for tracking

RECORD_SPEED: bool = False    # *  Record tracking speed

TRACKING_CHECK_WINDOW_LEN: int = 100  # Tracking algorithm backtracking window length

SMOOTH_WINDOW_LEN = 10  # Window length of smooth classification algorithm

PROB_THRESHOLD: float = 0.6  # Probability threshold of smooth classification algorithm

N_CLASS = 4    # *  Number of annotation categories

CLASS_NAME: List | None = ['G1', 'S', 'G2', 'M']   # *  Class name list
# CLASS_NAME = None

EXPORT_TRACKING_VISUALIZATION = True

INCLUDE_CELL_CYCLE = True

FILTER_THRESHOLD = 10